
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "egholed.h"
#include "bsmesh.h"
#include "bsfile.h"

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _bsf_ReadBCurve ( bsf_UserReaders *readers )
{
  void    *sp;
  char    *name;
  point4d *cp;
  byte    *mk;
  int     degree, spdimen;
  boolean rational;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( (readers->maxdeg+1)*sizeof(point4d) );
  mk = pkv_GetScratchMem ( readers->maxdeg+1 );
  if ( !name || !cp || !mk )
    goto failure;
  if ( !bsf_ReadBezierCurve4d ( readers->maxdeg, &degree, cp,
                                &spdimen, &rational, mk, name ) )
    goto failure;
  if ( readers->BezierCurveReader )
    (readers->BezierCurveReader)(readers->userData, name, degree, cp, spdimen, rational, mk );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBCurve*/

static boolean _bsf_ReadBPatch ( bsf_UserReaders *readers )
{
  void    *sp;
  char    *name;
  point4d *cp;
  byte    *mk;
  int     udeg, vdeg, pitch, spdimen;
  boolean rational;

  int maxdeg = readers->maxdeg;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( (maxdeg+1)*(maxdeg+1)*sizeof(point4d) );
  mk = pkv_GetScratchMem ( (maxdeg+1)*(maxdeg+1) );
  if ( !name || !cp || !mk )
    goto failure;
  if ( !bsf_ReadBezierPatch4d ( maxdeg, &udeg, &vdeg, &pitch, cp,
                                &spdimen, &rational, mk, name ) )
    goto failure;
  if ( readers->BezierPatchReader )
    (readers->BezierPatchReader)( readers->userData, name, udeg, vdeg, 
                                  pitch, cp, spdimen, rational, mk );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBPatch*/

static boolean _bsf_ReadBSCurve ( bsf_UserReaders *readers )
{
  void    *sp;
  char    *name;
  point4d *cp;
  byte    *mk;
  double  *knots;
  int     degree, lastknot, spdimen;
  boolean rational, closed;

  int maxlkn = readers->maxlkn;
  int maxdeg = readers->maxdeg;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( maxlkn*sizeof(point4d) );
  knots = pkv_GetScratchMemd ( maxlkn+1 );
  mk = pkv_GetScratchMem ( maxlkn );
  if ( !name || !cp || !knots || !mk )
    goto failure;
  if ( !bsf_ReadBSplineCurve4d ( maxdeg, maxlkn, maxlkn,
                                 &degree, &lastknot, knots, &closed, cp,
                                 &spdimen, &rational, mk, name ) )
    goto failure;
  if ( readers->BSplineCurveReader )
    (readers->BSplineCurveReader)( readers->userData, name, degree, lastknot,
                                   knots, closed, cp, spdimen, rational, mk );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSCurve*/

static boolean _bsf_ReadBSPatch ( bsf_UserReaders *readers )
{
  void    *sp;
  char    *name;
  point4d *cp;
  byte    *mk;
  double  *knotsu, *knotsv;
  int     spdimen, udeg, vdeg, lastknotu, lastknotv, pitch;
  boolean closed_u, closed_v, rational;

  int maxlkn = readers->maxlkn;
  int maxdeg = readers->maxdeg;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( maxlkn*maxlkn*sizeof(point4d) );
  knotsu = pkv_GetScratchMemd ( maxlkn );
  knotsv = pkv_GetScratchMemd ( maxlkn );
  mk = pkv_GetScratchMem ( maxlkn*maxlkn );
  if ( !name || !cp || !knotsu || !knotsv || !mk )
    goto failure;
  if ( !bsf_ReadBSplinePatch4d ( maxdeg, maxlkn,
                                 maxlkn*maxlkn,
                                 &udeg, &lastknotu, knotsu,
                                 &vdeg, &lastknotv, knotsv,
                                 &closed_u, &closed_v, &pitch, cp,
                                 &spdimen, &rational, mk, name ) )
    goto failure;
  if ( readers->BSplinePatchReader )
    (readers->BSplinePatchReader)( readers->userData,
               name, udeg, lastknotu, knotsu, vdeg, lastknotv, knotsv,
               closed_u, closed_v, pitch, cp, spdimen, rational, mk );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSPatch*/

static boolean _bsf_ReadBSMesh ( bsf_UserReaders *readers )
{
  void        *sp;
  char        *name;
  BSMvertex   *mv;
  BSMhalfedge *mhe;
  BSMfacet    *mfac;
  int         *mvhei, *mfhei;
  point4d     *vc;
  byte        *mkv;
  int         nv, nhe, nfac, spdimen, degree;
  boolean     rational;

  int maxnv = readers->maxnv;
  int maxnhe = readers->maxnhe;
  int maxnfac = readers->maxnfac;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  mv = pkv_GetScratchMem ( maxnv*sizeof(BSMvertex) );
  mhe = pkv_GetScratchMem ( maxnhe*sizeof(BSMhalfedge) );
  mfac = pkv_GetScratchMem ( maxnfac*sizeof(BSMfacet) );
  vc = pkv_GetScratchMem ( maxnv*sizeof(point4d) );
  mvhei = pkv_GetScratchMemi ( maxnhe );
  mfhei = pkv_GetScratchMemi ( maxnhe );
  mkv = pkv_GetScratchMem ( maxnv );
  if ( !name || !mv || !mhe || !mfac || !vc || !mvhei || !mfhei || !mkv )
    goto failure;
  if ( !bsf_ReadBSMesh4d ( maxnv, maxnhe, maxnfac,
                           &degree, &nv, mv, mvhei, vc,
                           &nhe, mhe, &nfac, mfac, mfhei, &spdimen, &rational,
                           mkv, name ) )
    goto failure;
  if ( readers->BSMeshReader )
    (readers->BSMeshReader)( readers->userData, name, degree, nv, mv, mvhei, 
                             vc, nhe, mhe, nfac, mfac, mfhei,
                             spdimen, rational, mkv );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSMesh*/

static boolean _bsf_ReadBSHole ( bsf_UserReaders *readers )
{
  void    *sp;
  int     hole_k, spdimen;
  char    *name;
  double  *knots;
  point2d *domcp;
  point4d *holecp;
  byte    *mk;
  boolean rational;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  knots = pkv_GetScratchMemd ( 11*GH_MAX_K );
  domcp = pkv_GetScratchMem ( (12*GH_MAX_K+1)*sizeof(point2d) );
  holecp = pkv_GetScratchMem ( (12*GH_MAX_K+1)*sizeof(point4d) );
  mk = pkv_GetScratchMem ( 12*GH_MAX_K+1 );
  if ( !knots || !domcp || !holecp || !mk )
    goto failure;
  if ( !bsf_ReadBSplineHoled ( GH_MAX_K, &hole_k, knots, domcp, holecp,
                               &spdimen, &rational, mk, name ) )
    goto failure;
  if ( readers->BSplineHoleReader )
    (readers->BSplineHoleReader)( readers->userData, name, hole_k, knots, 
                                  domcp, holecp, spdimen, rational, mk );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSHole*/

/* procedure of reading an entire file; for each item the application */
/* routine is called in order to enter the data into the application */
/* data structures */
boolean bsf_ReadBSFiled ( const char *filename, bsf_UserReaders *readers)
{
  void    *sp;

  sp = pkv_GetScratchMemTop ();
  if ( !bsf_OpenInputFile ( filename ) )
    return false;
  for (;;) {
    switch ( bsf_nextsymbol ) {
case BSF_SYMB_BCURVE:
      if ( !_bsf_ReadBCurve (readers) )
        goto failure;
      break;

case BSF_SYMB_BPATCH:
      if ( !_bsf_ReadBPatch (readers) )
        goto failure;
      break;

case BSF_SYMB_BSCURVE:
      if ( !_bsf_ReadBSCurve (readers) )
        goto failure;
      break;

case BSF_SYMB_BSPATCH:
      if ( !_bsf_ReadBSPatch (readers) )
        goto failure;
      break;

case BSF_SYMB_BSMESH:
      if ( !_bsf_ReadBSMesh (readers) )
        goto failure;
      break;

case BSF_SYMB_BSHOLE:
      if ( !_bsf_ReadBSHole (readers) )
        goto failure;
      break;

case BSF_SYMB_EOF:
      goto finish;

default:
      goto failure;
    }
  }
finish:
  bsf_CloseInputFile ();
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  bsf_CloseInputFile ();
  pkv_SetScratchMemTop ( sp );
  return false;
} /*bsf_ReadBSFiled*/


bsf_UserReaders * bsf_NewReaders(void) {
  bsf_UserReaders *readers = NULL; 
  PKV_MALLOC(readers, sizeof(bsf_UserReaders));
  if (!readers) {
    fprintf(stderr, "Cannot allocate memory for bsf_UserReader ... exiting");
    exit(11);
  }
  readers->userData = NULL;
  bsf_ClearReaders(readers);
  return readers;
}

void bsf_ClearReaders(bsf_UserReaders *readers) {
  if (readers) {
    readers->BezierCurveReader = NULL;
    readers->BSplineCurveReader = NULL;
    readers->BezierPatchReader = NULL;
    readers->BSplinePatchReader = NULL;
    readers->BSMeshReader = NULL;
    readers->BSplineHoleReader = NULL;
    if (readers->userData) {
      fprintf(stderr, 
         "Warning -- pointr to use data will be lost (bsf_ClearReaders()");
    } 
    readers->userData = NULL;
    readers->maxdeg = 10;
    readers->maxlkn = 100;
    readers->maxnv  = 1000;
    readers->maxnhe = 2000;
    readers->maxnfac = 1000;
  }
}
