
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#ifndef PKLINKAGE_H
#define PKLINKAGE_H

#ifndef PKVARIA_H
#include "pkvaria.h"
#endif
#ifndef PKNUM_H
#include "pknum.h"
#endif
#ifndef PKGEOM_H
#include "pkgeom.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* articulated object types */
#define PKKL_NONE          -1
#define PKKL_UNKNOWN        0
#define PKKL_BSPLINE_CURVE  1
#define PKKL_BSPLINE_PATCH  2
#define PKKL_BSPLINE_MESH   3

/* possible articulation methods */
#define PKKL_ART_OTHERHALF -1
#define PKKL_ART_NONE       0
#define PKKL_ART_TRANS_X    1
#define PKKL_ART_TRANS_Y    2
#define PKKL_ART_TRANS_Z    3
#define PKKL_ART_TRANS_V    4
#define PKKL_ART_SCALE_X    5
#define PKKL_ART_SCALE_Y    6
#define PKKL_ART_SCALE_Z    7
#define PKKL_ART_SCALE_XYZ  8
#define PKKL_ART_ROT_X      9
#define PKKL_ART_ROT_Y     10
#define PKKL_ART_ROT_Z     11
#define PKKL_ART_ROT_V     12
#define PKKL_ART_ROT_Q     13
#define PKKL_ART_AFFINE    14

#define PKKL_MAX_NAME_LENGTH 64

typedef struct pkkl_geom_obj {
    int     type;
    char    *name;
    char    *filename;
    int     nvc;                   /* number of vertex coordinates, 3 or 4 */
    int     nvert;                 /* number of vertices */
    double  *vert, *tvert;         /* pointers to vertices */
    trans3d Etr;                   /* extra object transformation */
    void    *data0, *data1;        /* pointers to object specific data */
    struct pkkl_geom_obj *nextobj; /* auxiliary object list pointer */
                                   /* object specific procedures */
    void    (*transform)(struct pkkl_geom_obj*,trans3d*,int,int*);
    void    (*postprocess)(struct pkkl_geom_obj*);
    void    (*redraw)(struct pkkl_geom_obj*,void*,void*);  /* drawing */
    void    (*destroy)(struct pkkl_geom_obj*);       /* deallocation */
  } pkkl_geom_obj;

typedef boolean (*pkkl_obj_init)(pkkl_geom_obj*,void*);
typedef void (*pkkl_obj_transform)(pkkl_geom_obj*,trans3d*,int,int*);
typedef void (*pkkl_obj_postprocess)(pkkl_geom_obj*);
typedef void (*pkkl_obj_redraw)(pkkl_geom_obj*,void*,void*);
typedef void (*pkkl_obj_destroy)(pkkl_geom_obj*);

typedef struct pkkl_obj_ref {
    int     on;     /* object number */
    int     nextr;  /* next reference */
    int     nv;     /* number of associated vertices */
    int     *vn;    /* vertex numbers */
  } pkkl_obj_ref;

typedef struct pkkl_link {
    char    *name;
    int     fref;  /* first object reference */
    int     fhj;   /* first half-joint */
    trans3d Mtr;   /* additional model transformation */
    char    tag;
  } pkkl_link;

typedef struct pkkl_halfjoint {
    int     l0, l1;     /* numbers of links - beginning and end */
    int     otherhalf;  /* number of the other half */
    int     nexthj;     /* next halfjoint */
    int     pnum;       /* number of articulation parameter */
    int     art;        /* type of articulation */
    trans3d Ftr;        /* front location transformation */
    trans3d Atr;        /* articulation transformation */
    trans3d Btr;        /* back location transformation */
  } pkkl_halfjoint;

typedef struct pkkl_linkage {
    int            maxobj,    nobj;     /* number of geometric objects */
    int            maxorefs,  norefs;   /* number of object references */
    int            maxlinks,  nlinks;   /* number of links */
    int            maxnhj,    nhj;      /* number of halfjoints */
    int            maxartpar, nartpar;  /* number of articulation parameters */
    pkkl_geom_obj  *obj;
    pkkl_obj_ref   *oref;
    pkkl_link      *link;
    pkkl_halfjoint *hj;
    double         *artp, *prevartp, *artprange;
    int            current_root;
    trans3d        current_root_tr;
    void           *usrdata;
  } pkkl_linkage;

pkkl_linkage *pkkl_NewLinkage ( int maxo, int maxl, int maxr,
                                int maxj, int maxp, double *artprange,
                                void *usrdata );
boolean pkkl_LinkageValid ( pkkl_linkage *linkage );
void pkkl_DestroyLinkage ( pkkl_linkage *linkage );

int pkkl_NewObject ( pkkl_linkage *linkage,
                     int type,
                     const char *name,
                     const char *filename,
                     int nvc, int nvert,
                     const trans3d *etrans,
                     void *data,
                     pkkl_obj_init init,
                     pkkl_obj_transform transform,
                     pkkl_obj_postprocess postprocess,
                     pkkl_obj_redraw redraw,
                     pkkl_obj_destroy destroy );
int pkkl_NewLink ( pkkl_linkage *linkage, const char *name );
int pkkl_NewObjRef ( pkkl_linkage *linkage, int lkn,
                     int on, int nv, int *vn );
int pkkl_NewJoint ( pkkl_linkage *linkage, int l0, int l1, int art, int pnum );
void pkkl_SetJointFtr ( pkkl_linkage *linkage, int jn,
                        trans3d *tr, boolean back );
void pkkl_SetJointBtr ( pkkl_linkage *linkage, int jn,
                        trans3d *tr, boolean front );

void pkkl_SetArtParam ( pkkl_linkage *linkage, int pno, int nump, double *par );
void pkkl_Articulate ( pkkl_linkage *linkage );
void pkkl_Redraw ( pkkl_linkage *linkage, void *appdata );

void pkkl_DefaultTransform ( pkkl_geom_obj *obj, trans3d *tr,
                             int nv, int *vn );

/* stub procedures */
boolean pkkl_obj_init_stub ( pkkl_geom_obj *obj, void *data );
void pkkl_obj_redraw_stub ( pkkl_geom_obj *obj, void *usrdata, void *appdata );
void pkkl_obj_stub ( pkkl_geom_obj *obj );

#ifdef __cplusplus
}
#endif

#endif /*PKLINKAGE_H*/

