
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#ifndef PKSCANNER_H
#define PKSCANNER_H

#ifndef _SYS_TIMES_H
#include <sys/times.h>
#endif

#ifndef PKVARIA_H
#include "pkvaria.h"
#endif
#ifndef AVLTREE_H
#include "avltree.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* options - bit masks */
#define PKS_OPT_C_OPERATORS  0x01
#define PKS_OPT_CPP_COMMENTS 0x02

/* symbol identifiers - somewhat random, to be rethought */
#define PKS_SYMB_ERROR      -3
#define PKS_SYMB_EOLN       -2
#define PKS_SYMB_EOF        -1
#define PKS_SYMB_NONE        0
#define PKS_SYMB_IDENT       1
#define PKS_SYMB_INTEGER     2
#define PKS_SYMB_FLOAT       3
#define PKS_SYMB_CHAR        4
#define PKS_SYMB_STRING      5
#define PKS_SYMB_PLUS        6  /* "+" */
#define PKS_SYMB_MINUS       7  /* "-" */
#define PKS_SYMB_STAR        8  /* "*" */
#define PKS_SYMB_SLASH       9  /* "/" */
#define PKS_SYMB_LPAREN     10  /* "(" */
#define PKS_SYMB_RPAREN     11  /* ")", previous + 1 */
#define PKS_SYMB_LBRACKET   12  /* "[" */
#define PKS_SYMB_RBRACKET   13  /* "]", previous + 1 */
#define PKS_SYMB_LBRACE     14  /* "{" */
#define PKS_SYMB_RBRACE     15  /* "}", previous + 1 */
#define PKS_SYMB_LANGLE     16  /* "<" */
#define PKS_SYMB_RANGLE     17  /* ">", previous + 1 */
#define PKS_SYMB_EQ         18  /* "=" */
#define PKS_SYMB_COMMA      19  /* "," */
#define PKS_SYMB_DOT        20  /* "." */
#define PKS_SYMB_SEMICOLON  21  /* ";" */
#define PKS_SYMB_COLON      22  /* ":" */
#define PKS_SYMB_PERCENT    23  /* "%" */
#define PKS_SYMB_HASH       24  /* "#" */
#define PKS_SYMB_EXCLAM     25  /* "!" */
#define PKS_SYMB_TILDE      26  /* "~" */
#define PKS_SYMB_AT         27  /* "@" */
#define PKS_SYMB_DOLLAR     28  /* "$" */
#define PKS_SYMB_ET         29  /* "&" */
#define PKS_SYMB_BACKSLASH  30  /* "\" */
#define PKS_SYMB_QUESTION   31  /* "?" */
#define PKS_SYMB_VERTBAR    32  /* "|" */
#define PKS_SYMB_WEDGE      33  /* "^" */
#define PKS_SYMB_EQEQ       34  /* "==" two-character symbols */
#define PKS_SYMB_LEQ        35  /* "<=" */
#define PKS_SYMB_GEQ        36  /* ">=", previous + 1 */
#define PKS_SYMB_NEQ        37  /* "!=" */
#define PKS_SYMB_ETET       38  /* "&&" */
#define PKS_SYMB_VERTBARBAR 39  /* "||" */
#define PKS_SYMB_PLUSPLUS   40  /* "++" */
#define PKS_SYMB_MINUSMINUS 41  /* "--" */
#define PKS_SYMB_LLANGLE    42  /* "<<" */
#define PKS_SYMB_RRANGLE    43  /* ">>", previous + 1 */
#define PKS_SYMB_RIGHTARROW 44  /* "->" */
#define PKS_SYMB_PLUSEQ     45  /* "+=" */
#define PKS_SYMB_MINUSEQ    46  /* "-=" */
#define PKS_SYMB_STAREQ     47  /* "*=" */
#define PKS_SYMB_SLASHEQ    48  /* "/=" */
#define PKS_SYMB_PERCENTEQ  49  /* "%=" */
#define PKS_SYMB_VERTBAREQ  50  /* "|=" */
#define PKS_SYMB_ETEQ       51  /* "&=" */
#define PKS_SYMB_WEDGEEQ    52  /* "^=" */
#define PKS_SYMB_LLANGLEEQ  53  /* "<<=" three-character symbols */
#define PKS_SYMB_RRANGLEEQ  54  /* ">>=" previous + 1 */
#define PKS_SYMB_OTHER      55
#define PKS_SYMB_MACROPAR   56
#define PKS_SYMB_BATCH_CMD  57  /* batch interpreter commands */
#define PKS_SYMB_BATCH_VAR  58  /* batch interpreter internal variable */
#define PKS_SYMB_FIRSTOTHER (PKS_SYMB_BATCH_VAR+1)

/* error codes */
#define PRERROR_NONE              0
#define PRERROR_INTERNAL          1
#define PRERROR_NO_MEMORY         2
#define PRERROR_BUFFER_OVERFLOW   3
#define PRERROR_WRONG_PARAMETER   4
#define PRERROR_MACRO_REDEF       5
#define PRERROR_PARAM_REDEF       6
#define PRERROR_WRONG_TOKEN       7
#define PRERROR_MACRO_UNDEF       8
#define PRERROR_RECURSIVE_EXP     9
#define PRERROR_PARAM_MATCH      10
#define PRERROR_IFSYNTAX         11
#define PRERROR_HASHIF           12
#define PRERROR_HASHINCLUDE      13
#define PRERROR_INPUT_FILE       14

/* maximal allowed line length */
#define PKS_SCANNER_MAX_LINE_LENGTH 1024

/* default buffer sizes */
#define PKS_SCANNER_BUF_LENGTH      4096
#define PKS_MACRO_BUF_LENGTH       65536

#define PKS_MAX_COMMENT_LENGTH         3

/* identifier space numbers */
#define ID_SPACE_NONE      0
#define ID_SPACE_CALC      1  /* cos, exp, sin etc. */
#define ID_SPACE_PREPROC   2  /* #define, #if, etc. */
#define ID_SPACE_BATCH     3  /* for, if, while etc. keywords */
#define ID_SPACE_BATCHVAR  4  /* loop control variables etc. */
#define ID_SPACE_BSFILE    5  /* libbsfile keywords */

#define pksbatch_MAXNLGT  15

typedef struct pks_file {
    FILE            *input;
    char            *filename;
    char            *inbuffer;
    int             inbufcount, inbufpos;
    int             linenum, colnum;
    short int       nextchar;
    int             ip;              /* storage for batch file compiler */
    struct pks_file *inclink;        /* file from which this was #included */
  } pks_file;

typedef struct pks_calcval {
    int      valtype;
    long int intval;
    double   floatval;
  } pks_calcval;

typedef struct pks_scanner {
    unsigned int options;
    pks_file     *infile;
    int          inbuflength, namebuflength;
    boolean      alloc_inb;   /* if true then allocate and free the input buffer */
    boolean      bkslnl;      /* if true then backslash cancels the following newline */
    byte         nrc;         /* number of characters returned, at most 3 */
    short int    rc[PKS_MAX_COMMENT_LENGTH]; /* buffer for the returned characters */
    short int    lc[PKS_MAX_COMMENT_LENGTH]; /* buffer for recent characters */
    char         bcomment1[PKS_MAX_COMMENT_LENGTH+1],
                 ecomment1[PKS_MAX_COMMENT_LENGTH+1];
    char         bcomment2[PKS_MAX_COMMENT_LENGTH+1],
                 ecomment2[PKS_MAX_COMMENT_LENGTH+1];
    char         *filename;
    int          nsline, nscolumn;
    short int    nextchar;
    short int    id_space;
    short int    nextsymbol;
    int          nextinteger;
    double       nextfloat;
    char         *nextname;
    int          namebufcount;
        /* preprocessor variables */
    AVLtree      macros;     /* table of macrodefinitions */
    void         *mstack;    /* stack of macros being expanded */
    short        ifsp;       /* #ifdef nesting level */
    short        error_code;
        /* scanner input/output virtual methods */
    int          (*InputData)(struct pks_scanner*);
    int          (*GetNextSymbol)(struct pks_scanner*);
        /* hooks for application identifiers */
    boolean      (*AppIdentFind)(struct pks_scanner*);
    boolean      (*AppIdentCalc)(struct pks_scanner*,boolean,pks_calcval*);
        /* hook for notifying #included files */
    boolean      (*IncludeFileHook)(struct pks_scanner*,boolean);
        /* application stuff */
    void         *libdata, *userdata;
  } pks_scanner;

typedef struct pks_batchexec {
    pks_scanner    fsc;   /* source file scanner/translator */
    pks_scanner    csc;   /* compiled code scanner/interpreter */
    clock_t        ticks;
    int            buflgt, codelgt;
    int            ip;
    char           *code;
    void           *aux;
    int            *var;
    boolean        fetch_next;
    boolean        (*AppIdentFind)(pks_scanner*);
    boolean        (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*);
  } pks_batchexec;

/* ////////////////////////////////////////////////////////////////////////// */
/* basic input text scanner - without any preprocessing */
boolean pks_InitScanner ( pks_scanner *sc,
                unsigned int options,
                int inbuflength, char *inbuffer, int maxnamelength,
                const char *bcomment, const char *ecomment,
                int (*InputData)(pks_scanner*),
                int (*GetNextSymbol)(pks_scanner*),
                boolean (*AppIdentFind)(pks_scanner*),
                boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                void *userdata );
void pks_GetNextChar ( pks_scanner *sc );
int pks_GetNextSymbol ( pks_scanner *sc );
void pks_ShutDownScanner ( pks_scanner *sc );

boolean pks_OpenInputFile ( pks_scanner *sc, const char *fn );
int pks_ReadInputFile ( pks_scanner *sc );
void pks_CloseInputFile ( pks_scanner *sc );

int pks_ReadEmptyString ( pks_scanner *sc );

int pks_BinSearchNameTable ( int nnames, int firstname, const char **names,
                             const char *name );

boolean pks_AppIdentFind ( pks_scanner *sc );
boolean pks_AppIdentCalc ( pks_scanner *sc, boolean ignore,
                           pks_calcval *value );

/* ////////////////////////////////////////////////////////////////////////// */
/* scanner with preprocessing */
boolean pks_InitPreprocessor ( pks_scanner *psc,
                unsigned int options,
                int inbuflength, int maxnamelength,
                const char *bcomment, const char *ecomment,
                int (*InputData)(pks_scanner*),
                int (*GetNextSymbol)(pks_scanner*),
                boolean (*AppIdentFind)(pks_scanner*),
                boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                void *userdata );
int pks_GetNextPSymbol ( pks_scanner *psc );
void pks_ShutDownPreprocessor ( pks_scanner *psc );

/* ////////////////////////////////////////////////////////////////////////// */
/* calculator */
    /* built in function names */
#define pkscalc_f_PI       0
#define pkscalc_f_ABS      1
#define pkscalc_f_ACOSH    2
#define pkscalc_f_ASINH    3
#define pkscalc_f_ATANH    4
#define pkscalc_f_ARCCOS   5
#define pkscalc_f_ARCSIN   6
#define pkscalc_f_ARCTAN   7
#define pkscalc_f_CBRT     8
#define pkscalc_f_CEILING  9
#define pkscalc_f_CLAMP   10
#define pkscalc_f_COS     11
#define pkscalc_f_COSH    12
#define pkscalc_f_EXP     13
#define pkscalc_f_FLOOR   14
#define pkscalc_f_LN      15
#define pkscalc_f_MAX     16
#define pkscalc_f_MIN     17
#define pkscalc_f_POW     18
#define pkscalc_f_ROUND   19
#define pkscalc_f_SIGN    20
#define pkscalc_f_SIN     21
#define pkscalc_f_SINH    22
#define pkscalc_f_SQRT    23
#define pkscalc_f_TAN     24
#define pkscalc_f_TANH    25
#define pkscalc_NUMNAMES (pkscalc_f_TANH+1)

extern const char *pkscalc_fnames[pkscalc_NUMNAMES];

int pkscalc_GetNextSymbol ( pks_scanner *sc );
boolean pkscalc_LogExpression ( pks_scanner *sc, pks_calcval *value );
boolean pkscalc_Expression ( pks_scanner *sc, pks_calcval *value );

/* ////////////////////////////////////////////////////////////////////////// */
/* batch file executor */
    /* keyword identifiers */
#define pksbatch_BREAK     0
#define pksbatch_CASE      1
#define pksbatch_CONTINUE  2
#define pksbatch_DEFAULT   3
#define pksbatch_DO        4
#define pksbatch_ELSE      5
#define pksbatch_FOR       6
#define pksbatch_GOTO      7
#define pksbatch_IF        8
#define pksbatch_LABEL     9
#define pksbatch_SWITCH   10
#define pksbatch_WHILE    11
#define pksbatch_NUMNAMES (pksbatch_WHILE+1)

boolean pksbatch_InitExecution ( pks_batchexec *batch,
                        unsigned int options,
                        const char *filename,
                        int inbuflength, int maxnamelength,
                        const char *bcomment, const char *ecomment,
                        char *params,
                        int (*InputData)(pks_scanner*),
                        boolean (*AppIdentFind)(pks_scanner*),
                        boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                        void *userdata );
void pksbatch_EndExecution ( pks_batchexec *batch );

void pksbatch_ListCompiledCode ( pks_batchexec *batch, FILE *f );
#ifdef __cplusplus
}
#endif

#endif

