
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadBSplinePatch4d ( bsf_Reader *reader,
                                 int maxdeg, int maxlastknot, int maxncpoints,
                                 int *udeg, int *lastknotu, double *knotsu,
                                 int *vdeg, int *lastknotv, double *knotsv,
                                 boolean *closed_u, boolean *closed_v,
                                 int *pitch, point4d *cpoints,
                                 int *spdimen, boolean *rational,
                                 char *name, int *ident )
{
  boolean _name, deg, knots_u, knots_v, c_points, dimen, _id, dep;
  int     ncpoints, dim, cpdimen;

  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_BSPATCH )  
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

        /* nothing has been read in yet */
  _name = deg = knots_u = knots_v = c_points = *rational =
  dimen = _id = dep = false;
  ncpoints = 0;
  if ( name )
    *name = 0;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_DEGREE:
        if ( deg )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadPatchDegree ( &reader->sc, maxdeg, udeg, vdeg ) )
          goto failure;
        deg = true;
        break;

  case BSF_KEY_DIM:
        if ( dimen )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadSpaceDim ( &reader->sc, 4, &dim ) )
          goto failure;
        dimen = true;
        break;

  case BSF_KEY_KNOTS_U:
        if ( knots_u )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadKnotSequenced ( &reader->sc, maxlastknot,
                                      lastknotu, knotsu, closed_u ) )
          goto failure;
        knots_u = true;
        break;

  case BSF_KEY_KNOTS_V:
        if ( knots_v )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadKnotSequenced ( &reader->sc, maxlastknot,
                                      lastknotv, knotsv, closed_v ) )
          goto failure;
        knots_v = true;
        break;

  case BSF_KEY_CPOINTS:
        if ( c_points )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        ncpoints = bsf_ReadPointsd ( &reader->sc, 4, maxncpoints,
                                     &cpoints[0].x, &cpdimen );
        c_points = true;
        if ( !ncpoints )
          goto failure;
        break;

  case BSF_KEY_RATIONAL:
        *rational = true;
        reader->sc.GetNextSymbol ( &reader->sc );
        break;

        /* trimmed patch domain */
  case BSF_KEY_TRIMMED:
        if ( !bsf_ReadTrimmedDomaind ( reader ) )
          goto failure;
        break;

        /* optional attributes */
  case BSF_KEY_CPOINTSMK:
        if ( !_bsf_ReadCPMark ( reader, maxncpoints ) )
          goto failure;
        break;

  case BSF_KEY_COLOR:
  case BSF_KEY_COLOUR:
        if ( !_bsf_ReadColour ( reader ) )
          goto failure;
        break;

        /* dependencies */
  case BSF_KEY_SPHERICAL_PRODUCT:
        if ( dep )
          goto failure;
        if ( !bsf_ReadDependencies ( reader ) )
          goto failure;
        dep = true;
        break;

  default:
        goto failure;  /* anything else is forbidden */
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( (*rational && dim < cpdimen-11) || (!*rational && dim < cpdimen) )
        goto failure;
      *spdimen = dim;
      if ( deg && knots_u && knots_v && c_points &&
           ncpoints == (*lastknotu-*udeg)*(*lastknotv-*vdeg) ) {
        *pitch = 4*(*lastknotv-*vdeg);
        return true;
      }
      else
        goto failure;

default:
      goto failure;  /* anything else is forbidden */
    }
  }

failure:
  return false;
} /*bsf_ReadBSplinePatch4d*/

