
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadBSplineHoled ( bsf_Reader *reader,
                               int maxk, int *hole_k, double *knots,
                               point2d *domain_cp, point4d *hole_cp,
                               int *spdimen, boolean *rational,
                               char *name, int *ident )
{
  boolean _name, sides, dimen, domcp, surfcp, _id;
  int     ns, nk, ncp, lkn, dim, cpdimen;

  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_BSHOLE )  
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

        /* nothing has ben read yet */
  _name = sides = dimen = domcp = surfcp = *rational = _id = false;
  if ( name )
    *name = 0;
  ns = nk = 0;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure; 
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_DIM:
        if ( dimen )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadSpaceDim ( &reader->sc, 4, spdimen ) )
          goto failure;
        dimen = true;
        break;

  case BSF_KEY_SIDES:
        if ( sides )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
          goto failure;
        if ( reader->sc.nextinteger < 3 || reader->sc.nextinteger > maxk )
          goto failure;
        *hole_k = ns = reader->sc.nextinteger;
        reader->sc.GetNextSymbol ( &reader->sc );
        sides = true;
        break;

  case BSF_KEY_KNOTS:
        if ( !sides )
          goto failure;
        if ( nk >= ns )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadKnotSequenced ( &reader->sc, 10, &lkn, &knots[nk*11], NULL ) )
          goto failure;
        if ( lkn != 10 )
          goto failure;
        nk ++;
        break;

  case BSF_KEY_DOMAIN:
        if ( domcp || !sides )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
             reader->sc.nextinteger != BSF_KEY_CPOINTS )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        ncp = bsf_ReadPointsd ( &reader->sc, 2, 12*ns+1, &domain_cp[0].x, &dim );
        if ( ncp != 12*ns+1 )
          goto failure;
        domcp = true;
        break;

  case BSF_KEY_CPOINTS:
        if ( surfcp || !sides )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        ncp = bsf_ReadPointsd ( &reader->sc, 4, 12*ns+1, &hole_cp[0].x, &cpdimen );
        if ( ncp != 12*ns+1 )
          goto failure;
        surfcp = true;
        break;

  case BSF_KEY_RATIONAL:
        reader->sc.GetNextSymbol ( &reader->sc );
        *rational = true;
        break;

        /* optional attributes */
  case BSF_KEY_CPOINTSMK:
        if ( !_bsf_ReadCPMark ( reader, 12*maxk+1 ) )
          goto failure;
        break;

  case BSF_KEY_COLOR:
  case BSF_KEY_COLOUR:
        if ( !_bsf_ReadColour ( reader ) )
          goto failure;
        break;

  default:
        goto failure;  /* anything else is forbidden */
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( (*rational && *spdimen < cpdimen-1) ||
           (!*rational && *spdimen < cpdimen) )
        goto failure;  
      if ( sides && domcp && surfcp && nk == ns )
        return true;
      else
        goto failure;

default:
      goto failure;  /* anything else is forbidden */
    }
  }

failure:
  return false;
} /*bsf_ReadBSplineHoled*/

