
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "bsmesh.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadBSMesh4d ( bsf_Reader *reader,
                           int maxnv, int maxnhe, int maxnfac,
                           int *degree,
                           int *nv, BSMvertex *mv, int *mvhei, point4d *vc,              
                           int *nhe, BSMhalfedge *mhe,      
                           int *nfac, BSMfacet *mfac, int *mfhei,
                           int *spdimen, boolean *rational,
                           char *name, int *ident )
{
  boolean _name, deg, vertices, halfedges, facets, dimen, _id, minus;
  int     nitems;
  int     i, k, d;

  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_BSMESH )  
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

        /* nothing has been read in yet */
  _name = deg = vertices = halfedges = facets = dimen = *rational = _id = false;
  if ( name )
    *name = 0;
  *degree = 0;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure; 
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_DIM:
        if ( dimen )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadSpaceDim ( &reader->sc, 4, spdimen ) )
          goto failure;
        dimen = true;
        break;

  case BSF_KEY_DEGREE:
        if ( deg )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_MINUS ) {
          minus = true;
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          minus = false;
        if ( !bsf_ReadCurveDegree ( &reader->sc, 255, degree ) )
          goto failure;
        if ( minus )
          *degree = -1;
        deg = true;
        break;

  case BSF_KEY_VERTICES:
        if ( vertices )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
          goto failure;
        nitems = reader->sc.nextinteger;
        if ( nitems < 0 || nitems > maxnv )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
          /* read in the vertices */
        k = 0;
        for ( i = 0; i < nitems; i++ ) {
          if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
            goto failure;
          d = 0;
          mv[i].firsthalfedge = k;
          do {
            reader->sc.GetNextSymbol ( &reader->sc );
            if ( k >= maxnhe || reader->sc.nextsymbol != PKS_SYMB_INTEGER )
              goto failure;
            mvhei[k++] = reader->sc.nextinteger;
            d ++;
            reader->sc.GetNextSymbol ( &reader->sc );
          } while ( reader->sc.nextsymbol == PKS_SYMB_COMMA );
          mv[i].degree = d;
          if ( !bsf_ReadPointd ( &reader->sc, 4, &vc[i].x, &d ) )
            goto failure;
          if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( i < nitems-1 ) {
            if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
              goto failure;
            reader->sc.GetNextSymbol ( &reader->sc );
          }
        }
        if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        *nv = nitems;
        vertices = true;
        break;

  case BSF_KEY_HALFEDGES:
        if ( halfedges )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
          goto failure;
        nitems = reader->sc.nextinteger;
        if ( nitems < 0 || nitems > maxnhe )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        /* read in the halfedges */
        for ( i = 0; i < nitems; i++ ) {
          if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
            goto failure;
          mhe[i].v0 = reader->sc.nextinteger;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
            goto failure;
          mhe[i].v1 = reader->sc.nextinteger;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
            goto failure;
          mhe[i].facetnum = reader->sc.nextinteger;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( !bsf_ReadIntNumber ( &reader->sc, &mhe[i].otherhalf ) )  /* this may be -1 or >= 0 */
            goto failure;
          if ( mhe[i].otherhalf < -1 )
            goto failure;
          if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( i < nitems-1 ) {
            if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
              goto failure;
            reader->sc.GetNextSymbol ( &reader->sc );
          }
        }
        if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        *nhe = nitems;
        halfedges = true;
        break;

  case BSF_KEY_FACETS:
        if ( facets )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
          goto failure;
        nitems = reader->sc.nextinteger;
        if ( nitems < 0 || nitems > maxnfac )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
          /* read in the facets */
        k = 0;
        for ( i = 0; i < nitems; i++ ) {
          if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
            goto failure;
          d = 0;
          mfac[i].firsthalfedge = k;
          do {
            reader->sc.GetNextSymbol ( &reader->sc );
            if ( k >= maxnhe || reader->sc.nextsymbol != PKS_SYMB_INTEGER )
              goto failure;
            mfhei[k++] = reader->sc.nextinteger;
            d ++;
            reader->sc.GetNextSymbol ( &reader->sc );
          } while ( reader->sc.nextsymbol == PKS_SYMB_COMMA );
          mfac[i].degree = d;
          if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
            goto failure;
          reader->sc.GetNextSymbol ( &reader->sc );
          if ( i < nitems-1 ) {
            if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
              goto failure;
            reader->sc.GetNextSymbol ( &reader->sc );
          }
        }
        if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        *nfac = nitems;
        facets = true;
        break;

  case BSF_KEY_RATIONAL:
        *rational = true;
        reader->sc.GetNextSymbol ( &reader->sc );
        break;

        /* optional attributes */
  case BSF_KEY_CPOINTSMK:
        if ( !_bsf_ReadCPMark ( reader, maxnv ) )
          goto failure;
        break;

  case BSF_KEY_HALFEDGEMK:
        if ( !_bsf_ReadHEdgeMark ( reader ) )
          goto failure;
        break;

  case BSF_KEY_FACETMK:
        if ( !_bsf_ReadFacetMark ( reader ) )
          goto failure;
        break;

  case BSF_KEY_COLOR:
  case BSF_KEY_COLOUR:
        if ( !_bsf_ReadColour ( reader ) )
          goto failure;
        break;

  default:
        goto failure;  /* anything else is forbidden */
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( vertices && halfedges && facets )
        return true;
      else
        goto failure;

default:
      goto failure;  /* anything else is forbidden */
    }
  }

failure:
  return false;
} /*bsf_ReadBSMesh4d*/

