
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h> 
#include <math.h> 
#include <malloc.h>
#include <string.h> 
#include <ctype.h>  

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadTrimmedDomaind ( bsf_Reader *reader )
{
  void    *sp;
  boolean one_contour;
  double  *knots;
  point4d *points;
  int     maxp, maxlkn, maxdeg, deg, dim, lkn, nvert;
  boolean closed, rational, in_contour;
  mbs_polycurved elem;

  sp = pkv_GetScratchMemTop ();
  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
         reader->sc.id_space != ID_SPACE_BSFILE ||
         reader->sc.nextinteger != BSF_KEY_TRIMMED )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_DOMAIN )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol == PKS_SYMB_LBRACE ) {
    one_contour = false;
    reader->sc.GetNextSymbol ( &reader->sc );
  }
  else
    one_contour = true;
  if ( reader->BeginReader )
    reader->BeginReader ( reader->userData, BSF_TRIMMED_DOMAIN );
  in_contour = true;
  for (;;) {
    elem.ident = -1;
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_BCURVE:
        maxdeg = reader->bc_maxdeg;
        maxp = maxdeg+1;
        points = pkv_GetScratchMem ( maxp*sizeof(point4d) );
        if ( !points )
          goto failure;
        elem.points = &points[0].x;
        if ( !bsf_ReadBezierCurve4d ( reader, maxdeg, &deg, points, &dim,
                                      &rational, NULL, &elem.ident ) )
          goto failure;
        if ( reader->TrimmedReader ) {
          elem.closing = reader->sc.nextsymbol == PKS_SYMB_RBRACE;
          if ( dim != 2 )
            goto failure;
          if ( rational )
            dim ++;
          elem.cpdimen = dim;
          pkv_Rearranged ( deg+1, dim, 4, dim, points );
          elem.degree = deg;
          elem.knots = NULL;
          elem.lastknot = -1;
          reader->TrimmedReader ( reader->userData, &elem );
        }
        pkv_SetScratchMemTop ( sp );
        break;

  case BSF_KEY_BSCURVE:
        maxdeg = reader->bsc_maxdeg;
        maxp = maxlkn = reader->bsc_maxlkn;
        points = pkv_GetScratchMem ( maxp*sizeof(point4d) );
        knots = elem.knots = pkv_GetScratchMemd ( maxlkn+1 );
        if ( !knots || !points )
          goto failure;
        elem.points = &points[0].x;
        if ( !bsf_ReadBSplineCurve4d ( reader, maxdeg, maxlkn, maxp,
                                       &deg, &lkn, knots,
                                       &closed, points, &dim,
                                       &rational, NULL, &elem.ident ) )
          goto failure;
        if ( reader->TrimmedReader ) {
          if ( closed )
            elem.closing = 2;
          else
            elem.closing = (reader->sc.nextsymbol == PKS_SYMB_RBRACE);
          if ( dim != 2 )
            goto failure;
          if ( rational )
            dim ++;
          elem.cpdimen = dim;
          pkv_Rearranged ( lkn-deg, dim, 4, dim, points );
          elem.degree = deg;
          elem.lastknot = lkn;
          reader->TrimmedReader ( reader->userData, &elem );
        }
        pkv_SetScratchMemTop ( sp );
        break;

  case BSF_KEY_POLYLINE:
        maxp = reader->poly_maxvert;
        points = pkv_GetScratchMem ( maxp*sizeof(point4d) );
        if ( !points )
          goto failure;
        elem.points = &points[0].x;
        if ( !bsf_ReadPolyline4d ( reader, maxp, &nvert, points, &dim,
                                   &rational, &closed, NULL, &elem.ident ) )
          goto failure;
        if ( reader->TrimmedReader ) {
          if ( closed )
            elem.closing = 2;
          else
            elem.closing = (reader->sc.nextsymbol == PKS_SYMB_RBRACE);
          if ( dim != 2 )
            goto failure;
          if ( rational )
            dim ++;
          elem.cpdimen = dim;
          pkv_Rearranged ( nvert, dim, 4, dim, points );
          elem.knots = NULL;
          elem.degree = 1;
          elem.lastknot = nvert-1;
          reader->TrimmedReader ( reader->userData, &elem );
        }
        pkv_SetScratchMemTop ( sp );
        break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_LBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( in_contour )     /* nesting too deep */
        goto failure;
      in_contour = true;    /* begin a next contour */
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      if ( one_contour || !in_contour ) {  /* end of the trimmed domain description */
        pkv_SetScratchMemTop ( sp );
        if ( reader->EndReader )
          reader->EndReader ( reader->userData, BSF_TRIMMED_DOMAIN, true );
        return true;
      }
      else {                /* end of a contour */
        in_contour = false;
        break;
      }

default:
      goto failure;
    }
  }

failure:
  pkv_SetScratchMemTop ( sp );
  if ( reader->EndReader )
    reader->EndReader ( reader->userData, BSF_TRIMMED_DOMAIN, false );
  return false;
} /*bsf_ReadTrimmedDomaind*/

