
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2005, 2014                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>  /* for testing */

#include "pkvaria.h"
#include "pkgeom.h"
#include "multibs.h"

#include "msgpool.h"

/* /////////////////////////////////////////// */
#define EPS 1.0e-12

boolean mbs_ClipBC2d ( int ncplanes, const vector3d *cplanes,
                       int degree, const point2d *cpoints,
                       boolean (*output) (int degree, const point2d *cpoints) )
{
  void    *sp;
  point2d *p, *q;
  double   *a, *t, x;
  int     i, j, nt, nz;
  boolean visible;

  sp = pkv_GetScratchMemTop ();

  t = pkv_GetScratchMemd ( ncplanes*degree + 2 );
  if ( t ) {
    t[0] = 0.0;
    nt = 1;

    for ( j = 0; j < ncplanes; j++ ) {
        /* compute the polynomial coefficients */
      if ( !(a = pkv_GetScratchMemd ( degree+1)) )
        goto failure;
      for ( i = 0; i <= degree; i++ )
        a[i] = DotProduct2d ( &cpoints[i], (point2d*)&cplanes[j] ) +
               cplanes[j].z;

      if ( !mbs_FindPolynomialZerosd ( degree, a, &nz, &t[nt], EPS ) )
        goto failure;
      nt += nz;
    }
    t[nt] = 1.0;
    pkv_SortFast ( sizeof(double), ID_IEEE754_DOUBLE, sizeof(double),
                   0, nt+1,  t );
        /* now output the arcs */
    p = pkv_GetScratchMem ( 2*(degree+1)*sizeof(point2d) );
    if ( !p )
      goto failure;
    q = &p[degree+1];
    for ( i = 0; i < nt; i++ ) {
          /* is it visible? */
      x = 0.5*(t[i]+t[i+1]);
      mbs_BCHornerC2d ( degree, cpoints, x, p );
      for ( j = 0, visible = true;  j < ncplanes && visible;  j++ ) {
        x = DotProduct2d ( p, (vector2d*)&cplanes[j] ) + cplanes[j].z;
        visible = (boolean)(x > 0.0);
      }
      if ( visible ) {
        memcpy ( p, cpoints, (degree+1)*sizeof(point2d) );
        mbs_DivideBC2d ( degree, t[i], p, q );
        mbs_DivideBC2d ( degree, (t[i+1]-t[i])/(1.0-t[i]), p, q );
        if ( !output ( degree, q ) )
          goto failure;
      }
    }
  }
  else
    goto failure;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_ClipBC2d*/

boolean mbs_ClipBC2Rd ( int ncplanes, const vector3d *cplanes,
                        int degree, const point3d *cpoints,
                        boolean (*output) (int degree, const point3d *cpoints) )
{
  void    *sp;
  point3d *p, *q;
  double   *a, *t, x;
  int     i, j, nt, nz;
  boolean visible;

  sp = pkv_GetScratchMemTop ();

  t = pkv_GetScratchMemd ( ncplanes*degree + 2 );
  if ( t ) {
    t[0] = 0.0;
    nt = 1;

    for ( j = 0; j < ncplanes; j++ ) {
        /* compute the polynomial coefficients */
      if ( !(a = pkv_GetScratchMemd ( degree+1)) )
        goto failure;
      for ( i = 0; i <= degree; i++ )
        a[i] = DotProduct3d ( &cpoints[i], &cplanes[j] );

      if ( !mbs_FindPolynomialZerosd ( degree, a, &nz, &t[nt], EPS ) )
        goto failure;
      nt += nz;
    }
    t[nt] = 1.0;
    pkv_SortFast ( sizeof(double), ID_IEEE754_DOUBLE, sizeof(double),
                   0, nt+1,  t );
        /* now output the arcs */
    p = pkv_GetScratchMem ( 2*(degree+1)*sizeof(point3d) );
    if ( !p )
      goto failure;
    q = &p[degree+1];
    for ( i = 0; i < nt; i++ ) {
          /* is it visible? */
      x = 0.5*(t[i]+t[i+1]);
      mbs_BCHornerC3d ( degree, cpoints, x, p );
      for ( j = 0, visible = true;  j < ncplanes && visible;  j++ ) {
        x = DotProduct3d ( p, &cplanes[j] );
        visible = (boolean)(x > 0.0);
      }
      if ( visible ) {
        memcpy ( p, cpoints, (degree+1)*sizeof(point3d) );
        mbs_DivideBC3d ( degree, t[i], p, q );
        mbs_DivideBC3d ( degree, (t[i+1]-t[i])/(1.0-t[i]), p, q );
        if ( !output ( degree, q ) )
          goto failure;
      }
    }
  }
  else
    goto failure;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_ClipBC2Rd*/

boolean mbs_ClipBC3d ( int ncplanes, const vector4d *cplanes,
                       int degree, const point3d *cpoints,
                       boolean (*output) (int degree, const point3d *cpoints) )
{
  void    *sp;
  point3d *p, *q;
  double   *a, *t, x;
  int     i, j, nt, nz;
  boolean visible;

  sp = pkv_GetScratchMemTop ();

  t = pkv_GetScratchMemd ( ncplanes*degree + 2 );
  if ( t ) {
    t[0] = 0.0;
    nt = 1;

    for ( j = 0; j < ncplanes; j++ ) {
        /* compute the polynomial coefficients */
      if ( !(a = pkv_GetScratchMemd ( degree+1)) )
        goto failure;
      for ( i = 0; i <= degree; i++ )
        a[i] = DotProduct3d ( &cpoints[i], (point3d*)&cplanes[j] ) +
               cplanes[j].w;

      if ( !mbs_FindPolynomialZerosd ( degree, a, &nz, &t[nt], EPS ) )
        goto failure;
      nt += nz;
    }
    t[nt] = 1.0;
    pkv_SortFast ( sizeof(double), ID_IEEE754_DOUBLE, sizeof(double),
                   0, nt+1,  t );
        /* now output the arcs */
    p = pkv_GetScratchMem ( 2*(degree+1)*sizeof(point3d) );
    if ( !p )
      goto failure;
    q = &p[degree+1];
    for ( i = 0; i < nt; i++ ) {
          /* is it visible? */
      x = 0.5*(t[i]+t[i+1]);
      mbs_BCHornerC3d ( degree, cpoints, x, p );
      for ( j = 0, visible = true;  j < ncplanes && visible;  j++ ) {
        x = DotProduct3d ( p, (vector3d*)&cplanes[j] ) + cplanes[j].w;
        visible = (boolean)(x > 0.0);
      }
      if ( visible ) {
        memcpy ( p, cpoints, (degree+1)*sizeof(point3d) );
        mbs_DivideBC3d ( degree, t[i], p, q );
        mbs_DivideBC3d ( degree, (t[i+1]-t[i])/(1.0-t[i]), p, q );
        if ( !output ( degree, q ) )
          goto failure;
      }
    }
  }
  else
    goto failure;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_ClipBC3d*/

boolean mbs_ClipBC3Rd ( int ncplanes, const vector4d *cplanes,
                        int degree, const point4d *cpoints,
                        boolean (*output) (int degree, const point4d *cpoints) )
{
  void     *sp;
  point4d  *p, *q;
  double   *a, *t, x;
  int      i, j, nt, nz;
  boolean  visible;

  sp = pkv_GetScratchMemTop ();

  t = pkv_GetScratchMemd ( ncplanes*degree + 2 );
  if ( t ) {
    t[0] = 0.0;
    nt = 1;

    for ( j = 0; j < ncplanes; j++ ) {
        /* compute the polynomial coefficients */
      if ( !(a = pkv_GetScratchMemd ( degree+1)) )
        goto failure;
      for ( i = 0; i <= degree; i++ )
        a[i] = DotProduct4d ( &cpoints[i], &cplanes[j] );

      if ( !mbs_FindPolynomialZerosd ( degree, a, &nz, &t[nt], EPS ) )
        goto failure;
      nt += nz;
    }
    t[nt] = 1.0;
    pkv_SortFast ( sizeof(double), ID_IEEE754_DOUBLE, sizeof(double),
                   0, nt+1,  t );
        /* now output the arcs */
    p = pkv_GetScratchMem ( 2*(degree+1)*sizeof(point4d) );
    if ( !p )
      goto failure;
    q = &p[degree+1];
    for ( i = 0; i < nt; i++ ) {
          /* is it visible? */
      x = 0.5*(t[i]+t[i+1]);
      mbs_BCHornerC4d ( degree, cpoints, x, p );
      for ( j = 0, visible = true;  j < ncplanes && visible;  j++ ) {
        x = DotProduct4d ( p, &cplanes[j] );
        visible = (boolean)(x > 0.0);
      }
      if ( visible ) {
        memcpy ( p, cpoints, (degree+1)*sizeof(point4d) );
        mbs_DivideBC4d ( degree, t[i], p, q );
        mbs_DivideBC4d ( degree, (t[i+1]-t[i])/(1.0-t[i]), p, q );
        if ( !output ( degree, q ) )
          goto failure;
      }
    }
  }
  else
    goto failure;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_ClipBC3Rd*/

