
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <limits.h>
#include <string.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "multibs.h"


typedef struct {
    Box2f box;
    float t0, t1;
    float *cp;  /* control points */
  } _mbs_arc;

static void _mbs_FindACPBox2f ( int dim, int deg, float *cp, Box2f *box )
{
  int   i, j;
  float u;

  if ( dim == 2 ) {
    box->x0 = box->x1 = cp[0];
    box->y0 = box->y1 = cp[1];
    for ( i = 1, j = 2;  i <= deg;  i++, j += 2 ) {
      if ( cp[j] < box->x0 )        box->x0 = cp[j];
      else if ( cp[j] > box->x1 )   box->x1 = cp[j];
      if ( cp[j+1] < box->y0 )      box->y0 = cp[j+1];
      else if ( cp[j+1] > box->y1 ) box->y1 = cp[j+1];
    }
  }
  else { /* dim == 3, homogeneous coordinates in use  */
    box->x0 = box->x1 = cp[0]/cp[2];
    box->y0 = box->y1 = cp[1]/cp[2];
    for ( i = 1, j = 3;  i <= deg;  i++, j += 3 ) {
      u = cp[j]/cp[j+2];
      if ( u < box->x0 )      box->x0 = u;
      else if ( u > box->x1 ) box->x1 = u;
      u = cp[j+1]/cp[j+2];
      if ( u < box->y0 )      box->y0 = u;
      else if ( u > box->y1 ) box->y1 = u;
    }
  }
} /*_mbs_FindACPBox2f*/

static int _mbs_PointLineSegmentIndexf ( short int dim, int lkn, float *points,
                                         float x, float y )
{
  int     index;
  char    sx0, sy0, sx1, sy1;
  point2f p0, p1;
  float   xt;
  int     i;

  index = 0;
  if ( dim == 2 ) {
    memcpy ( &p0, points, sizeof(point2f) );
    sx0 = pkv_signf ( p0.x-x );
    sy0 = pkv_signf ( p0.y-y );
    for ( i = 1; i <= lkn; i++ ) {
      memcpy ( &p1, &points[2*i], sizeof(point2f) );
      sx1 = pkv_signf ( p1.x-x );
      sy1 = pkv_signf ( p1.y-y );
      if ( (sx0 >= 0 || sx1 >= 0) && sy0 != sy1 ) {
        xt = p0.x + (y-p0.y)/(p1.y-p0.y)*(p1.x-p0.x);
        if ( xt >= x ) {
          if ( sy0 <= 0 && sy1 > 0 )      index ++;
          else if ( sy0 > 0 && sy1 <= 0 ) index --;
        }
      }
      p0 = p1;  sx0 = sx1;  sy0 = sy1;
    }
  }
  else { /* dim == 3, homogeneous coordinates in use */
    SetPoint2f ( &p0, points[0]/points[2], points[1]/points[2] );
    sx0 = pkv_signf ( p0.x-x );
    sy0 = pkv_signf ( p0.y-y );
    for ( i = 1; i <= lkn; i++ ) {
      SetPoint2f ( &p1, points[3*i]/points[3*i+2], points[3*i+1]/points[3*i+2] );
      sx1 = pkv_signf ( p1.x-x );
      sy1 = pkv_signf ( p1.y-y );
      if ( (sx0 >= 0 || sx1 >= 0) && sy0 != sy1 ) {
        xt = p0.x + (y-p0.y)/(p1.y-p0.y);
        if ( xt >= x ) {
          if ( sy0 <= 0 && sy1 > 0 )      index ++;
          else if ( sy0 > 0 && sy1 <= 0 ) index --;
        }
      }
      p0 = p1;  sx0 = sx1;  sy0 = sy1;
    }
  }
  return index;
} /*_mbs_PointLineSegmentIndexf*/

static int _mbs_PointBezierArcIndexf ( int dim, int deg, float *cpoints,
                                       float x, float y,
                                       float *workspace )
{
  int           index;
  int           sp;
  int           size;
  _mbs_arc      stk[MBS_INDEX_MAXLEVELF];
  float         *p, *cp, t;
  _mbs_BezPolyf bpoly;
  int           i, j;
  char          s0, s1;
  boolean       nonzero, error;
  point2f       q, ls[2];

  index = 0;
  size = dim*(deg+1);
  bpoly.deg = deg;
  bpoly.coeff = p = workspace;
  stk[0].cp = &p[deg+1];
  memcpy ( stk[0].cp, cpoints, size*sizeof(float) );
  stk[0].t0 = 0.0;  stk[0].t1 = 1.0;
  sp = 1;
  do {
    sp --;
    cp = stk[sp].cp;
    _mbs_FindACPBox2f ( dim, deg, cp, &stk[sp].box );
    if ( stk[sp].box.x1 >= x && stk[sp].box.y0 <= y && stk[sp].box.y1 > y ) {
      if ( stk[sp].box.x0 >= x ) {
        /* the convex hull suffices */
        if ( dim == 2 ) {
          s0 = pkv_signf ( cp[1] - y );
          s1 = pkv_signf ( cp[2*deg+1] - y );
        }
        else {
          s0 = pkv_signf ( cp[1]/cp[2] - y );
          s1 = pkv_signf ( cp[3*deg+1]/cp[3*deg+2] - y );
        }
        if ( s0 <= 0 && s1 > 0 )      index ++;
        else if ( s0 > 0 && s1 <= 0 ) index --;
          /* else the index must be unchanged */
      }
      else {
        /* find the polynomial coefficients */
        if ( dim == 2 ) {
          for ( i = 0, j = 1;  i <= deg;  i++, j += 2 )
            p[i] = cp[j] - y;
        }
        else {
          for ( i = 0, j = 1;  i <= deg;  i++, j += 3 )
            p[i] = cp[j] - y*cp[j+1];
        }
        if ( pkn_OneSignChangef ( deg+1, p, &nonzero ) ) {
          if ( nonzero ) {
            t = pkn_Illinoisf ( _mbs_EvalBezPolyf, (void*)&bpoly, 0.0, 1.0,
                                MBS_INDEX_EPSF, &error );
            t = stk[sp].t0 + t*(stk[sp].t1-stk[sp].t0);
            if ( dim == 2 )
              mbs_BCHornerC2f ( deg, cpoints, t, &q );
            else
              mbs_BCHornerC2Rf ( deg, (point3f*)cpoints, t, &q );
            if ( q.x >= x ) {
              if ( p[0] < p[deg] )
                index ++;
              else
                index --;
            }
          }
        }
        else if ( sp < MBS_INDEX_MAXLEVELF &&
                  stk[sp].t1-stk[sp].t0 > MBS_INDEX_EPSF ) {  /* divide the arc */
          stk[sp+1].cp = &stk[sp].cp[size];
          mbs_multiBisectBezCurvesf ( deg, 1, dim, 0, stk[sp].cp, stk[sp+1].cp );
          stk[sp+1].t0 = stk[sp].t0;
          stk[sp+1].t1 = stk[sp].t0 = 0.5*(stk[sp].t0+stk[sp].t1);
          sp += 2;
        }
        else {  /* replace the arc by a line segment - */
                /* it's the best that may be done */
          if ( dim == 2 ) {
            memcpy ( &ls[0], &stk[sp].cp[0], sizeof(point2f) );
            memcpy ( &ls[1], &stk[sp].cp[2*deg], sizeof(point2f) );
          }
          else {
            Point3to2f ( (point3f*)&stk[sp].cp[0], &ls[0] );
            Point3to2f ( (point3f*)&stk[sp].cp[3*deg], &ls[1]  );
          }
          index += _mbs_PointLineSegmentIndexf ( 2, 1, &ls[0].x, x, y );
        }
      }
    }
  } while ( sp > 0 );
  return index;
} /*_mbs_PointBezierArcIndexf*/

int mbs_PointPolycurveIndexf ( int nelem, mbs_polycurvef *poly,
                               float x, float y,
                               float *workspace )
{
#define CONVERTP(dim,p,q) \
  { if ( dim == 2 ) memcpy ( q, p, sizeof(point2f) ); \
    else Point3to2f ( (point3f*)p, q ); }

  int     index;
  float   *points, *bsbuf;
  int     dim, deg, lkn, ku;
  int     i, j;
  boolean first;
  point2f segment[3];

  index = 0;
  first = true;
  for ( i = 0; i < nelem; i++ ) {
    dim = poly[i].cpdimen;
    deg = poly[i].degree;
    lkn = poly[i].lastknot;
    points = poly[i].points;
    if ( poly[i].knots ) { /* a B-spline curve */
      ku = mbs_NumKnotIntervalsf ( deg, lkn, poly[i].knots );
      bsbuf = pkv_GetScratchMemf ( dim*(deg+1)*ku );
      if ( !bsbuf )
        return INT_MIN;
      if ( !mbs_multiBSCurvesToBezf ( dim, 1, deg, lkn, poly[i].knots,
                                      0, points, &ku, NULL, NULL, 0, bsbuf ) )
        return INT_MIN;
      CONVERTP ( dim, bsbuf, &segment[1] )
      if ( first ) segment[2] = segment[1];
        else index += _mbs_PointLineSegmentIndexf ( 2, 1, &segment[0].x, x, y );
      for ( j = 0; j < ku; j++ )
        index += _mbs_PointBezierArcIndexf ( dim, deg, &bsbuf[j*dim*(deg+1)],
                                             x, y, workspace );
      CONVERTP ( dim, &bsbuf[ku*dim*(deg+1)-dim], segment )
      pkv_SetScratchMemTop ( bsbuf );
    }
    else if ( lkn == -1 ) { /* a Bezier curve */
      CONVERTP ( dim, points, &segment[1] )
      if ( first ) segment[2] = segment[1];
        else index += _mbs_PointLineSegmentIndexf ( 2, 1, &segment[0].x, x, y );
      if ( deg == 1 )
        index += _mbs_PointLineSegmentIndexf ( dim, 1, points, x, y );
      else
        index += _mbs_PointBezierArcIndexf ( dim, deg, points, x, y, workspace );
      CONVERTP ( dim, &points[dim*deg], segment );
    }
    else { /* a polyline made of lkn line segments */
      CONVERTP ( dim, points, &segment[1] )
      if ( first ) segment[2] = segment[1];
        else index += _mbs_PointLineSegmentIndexf ( 2, 1, &segment[0].x, x, y );
      if ( first ) segment[2] = segment[1];
      index += _mbs_PointLineSegmentIndexf ( dim, lkn, points, x, y );
      CONVERTP ( dim, &points[dim*lkn], segment )
    }
    if ( (first = (poly[i].closing != 0)) ) {
      segment[1] = segment[2];
      index += _mbs_PointLineSegmentIndexf ( 2, 1, &segment[0].x, x, y );
    }
  }
  return index;
#undef CONVERTP
} /*mbs_PointPolycurveIndexf*/

