
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <limits.h>
#include <string.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "multibs.h"

#define EPS 1.0e-12

boolean mbs_FindBC2ClosestPointd ( int degree, const point2d *cp,
                                   const point2d *q,
                                   double *t, point2d *p, double *dist )
{
  int      n, i, j, nz;
  double   *c, *z, d, dmin;
  vector2d *a, *b, v;
  point2d  s;
  
  n = 2*degree-1;
  c = pkv_GetScratchMemd ( 6*degree+3 );
  if ( !c )
    return false;
  a = (vector2d*)&c[n+1];
  b = &a[degree+1];
        /* compute the polynomial coefficients */
  for ( i = 0; i <= degree; i++ )
    SubtractPoints2d ( q, &cp[i], &a[i] );
  dmin = DotProduct2d ( a, a );  *t = 0.0;  *p = cp[0];
  d = DotProduct2d ( &a[degree], &a[degree] );
  if ( d < dmin ) { dmin = d;  *t = 1.0; *p = cp[degree]; }
  mbs_multiBezScaled ( degree, 1, 1, 2, 0, &a[0].x );
  for ( i = 0; i < degree; i++ )
    SubtractPoints2d ( &cp[i+1], &cp[i], &b[i] );
  mbs_multiBezScaled ( degree-1, 1, 1, 2, 0, &b[0].x );
  memset ( c, 0, (n+1)*sizeof(double) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < degree; j++ )
      c[i+j] += DotProduct2d ( &a[i], &b[j] );
  mbs_multiBezUnscaled ( n, 1, 1, 1, 0, c );
  z = &a[0].x;
        /* find the closest point */
  if ( mbs_FindPolynomialZerosd ( n, c, &nz, z, EPS ) ) {
    for ( i = 0; i < nz; i++ ) {
      mbs_BCHornerC2d ( degree, cp, z[i], &s );
      SubtractPoints2d ( q, &s, &v );
      d = DotProduct2d ( &v, &v );
      if ( d < dmin )
        { dmin = d;  *t = z[i];  *p = s; }
    }
  }
  *dist = sqrt ( dmin );
  pkv_SetScratchMemTop ( c );
  return true;
} /*mbs_FindBC2ClosestPointd*/

boolean mbs_FindBS2ClosestPointd ( int degree, int lkn, const double *knots,
                                   const point2d *cpoints,
                                   const point2d *q,
                                   double *t, point2d *p, double *dist )
{
  void    *sp;
  int     ku, lknu;
  double  *kn, s, d;
  point2d *bcp, r;
  int     i;

  sp = pkv_GetScratchMemTop ();
  ku = mbs_NumKnotIntervalsd ( degree, lkn, knots );
  lknu = (degree+1)*(ku+1)-1;
  kn = pkv_GetScratchMemd ( lknu+1 );
  bcp = pkv_GetScratchMem ( (lknu-degree)*sizeof(point2d) );
  if ( !kn || !bcp )
    goto failure;
  if ( !mbs_BSToBezC2d ( degree, lkn, knots, cpoints, &ku, &lknu, kn, bcp ) )
    goto failure;
  if ( !mbs_FindBC2ClosestPointd ( degree, bcp, q, &s, p, dist ) )
    goto failure;
  *t = kn[degree] + s*(kn[degree+1]-kn[degree]);
  for ( i = 1; i < ku; i++ ) {
    if ( !mbs_FindBC2ClosestPointd ( degree, &bcp[i*(degree+1)], q,
                                     &s, &r, &d ) )
      goto failure;
    if ( d < *dist ) {
      *t = kn[(i+1)*(degree+1)-1] + s*(kn[(i+1)*(degree+1)]-kn[(i+1)*(degree+1)-1]);
      *p = r;
      *dist = d;
    }
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_FindBS2ClosestPointd*/

boolean mbs_FindBC2RClosestPointd ( int degree, const point3d *cp,
                                    const point2d *q,
                                    double *t, point2d *p, double *dist )
{
  int      n, i, j, nz;
  double   *c, *z, d, dmin;
  vector2d *a, *e, v;
  vector3d *r, *b;
  point2d  s;
  
  n = 3*degree-1;
  c = pkv_GetScratchMemd ( 15*degree+5 );
  if ( !c )
    return false;
  a = (vector2d*)&c[n+1];
  b = (vector3d*)&a[degree+1];
  r = &b[degree];
  e = (vector2d*)&r[degree+1];
        /* compute the polynomial coefficients */
  memcpy ( r, cp, (degree+1)*sizeof(point3d) );
  for ( i = 0; i <= degree; i++ )
    SetVector2d ( &a[i], cp[i].z*q->x-cp[i].x, cp[i].z*q->y-cp[i].y );
  mbs_multiBezScaled ( degree, 1, 1, 3, 0, &r[0].x );
  mbs_multiBezScaled ( degree, 1, 1, 2, 0, &a[0].x );
  for ( i = 0; i < degree; i++ )
    SubtractPoints3d ( &cp[i+1], &cp[i], &b[i] );
  mbs_multiBezScaled ( degree-1, 1, 1, 3, 0, &b[0].x );
  memset ( e, 0, 4*degree*sizeof(double) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < degree; j++ ) {
      e[i+j].x += r[i].z*b[j].x-r[i].x*b[j].z;
      e[i+j].y += r[i].z*b[j].y-r[i].y*b[j].z;
    }
  memset ( c, 0, (n+1)*sizeof(double) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < 2*degree; j++ )
      c[i+j] += DotProduct2d ( &a[i], &e[j] );
  mbs_multiBezUnscaled ( n, 1, 1, 1, 0, c );
  z = &a[0].x;
        /* find the closest point */
  Point3to2d ( &cp[0], &s );
  SubtractPoints2d ( q, &s, &v );
  dmin = DotProduct2d ( &v, &v );  *t = 0.0;  *p = s;
  Point3to2d ( &cp[degree], &s );
  SubtractPoints2d ( q, &s, &v );
  d = DotProduct2d ( &v, &v );
  if ( d < dmin )
    { dmin = d;  *t = 1.0;  *p = s; }
  if ( mbs_FindPolynomialZerosd ( n, c, &nz, z, EPS ) ) {
    for ( i = 0; i < nz; i++ ) {
      mbs_BCHornerC2Rd ( degree, cp, z[i], &s );
      SubtractPoints2d ( q, &s, &v );
      d = DotProduct2d ( &v, &v );
      if ( d < dmin )
        { dmin = d;  *t = z[i];  *p = s; }
    }
  }
  *dist = sqrt ( dmin );
  pkv_SetScratchMemTop ( c );
  return true;
} /*mbs_FindBC2RClosestPointd*/

boolean mbs_FindBS2RClosestPointd ( int degree, int lkn, const double *knots,
                                    const point3d *cpoints,
                                    const point2d *q,
                                    double *t, point2d *p, double *dist )
{
  void    *sp;
  int     ku, lknu;
  double  *kn, s, d;
  point3d *bcp;
  point2d r;
  int     i;

  sp = pkv_GetScratchMemTop ();
  ku = mbs_NumKnotIntervalsd ( degree, lkn, knots );
  lknu = (degree+1)*(ku+1)-1;
  kn = pkv_GetScratchMemd ( lknu+1 );
  bcp = pkv_GetScratchMem ( (lknu-degree)*sizeof(point3d) );
  if ( !kn || !bcp )
    goto failure;
  if ( !mbs_BSToBezC3d ( degree, lkn, knots, cpoints, &ku, &lknu, kn, bcp ) )
    goto failure;
  if ( !mbs_FindBC2RClosestPointd ( degree, bcp, q, &s, p, dist ) )
    goto failure;
  *t = kn[degree] + s*(kn[degree+1]-kn[degree]);
  for ( i = 1; i < ku; i++ ) {
    if ( !mbs_FindBC2RClosestPointd ( degree, &bcp[i*(degree+1)], q,
                                      &s, &r, &d ) )
      goto failure;
    if ( d < *dist ) {
      *t = kn[(i+1)*(degree+1)-1] + s*(kn[(i+1)*(degree+1)]-kn[(i+1)*(degree+1)-1]);
      *p = r;
      *dist = d;
    }
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
}  /*mbs_FindBS2RClosestPointd*/

