
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2019                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <math.h>
#include <string.h>
#include <stdio.h>  /* for testing */

#include "pkvaria.h"
#include "pkgeom.h"
#include "multibs.h"


/* ///////////////////////////////////////////// */
/* computing values of B-spline functions and    */
/* their first and second order derivatives at t */
/* fnz - index of the first nonzero function,    */
/* whose value is stored in bfv[0].              */

void mbs_deBoorBasisDer2d ( int degree, int lastknot, const double *knots,  
                            double t, int *fnz, int *nnz,
                            double *bfv, double *dbfv, double *ddbfv )
{
  int i, j, k, l, r;
  double alpha, beta, gamma, delta, eps;
  double *auxkn, *auxbfv, *auxdbfv, *auxddbfv;
  int   auxsize;

  if ( degree < 2 )
    goto way_out;
  k = mbs_FindKnotIntervald ( -1, lastknot, knots, t, &r );

                     /* first, there are two special cases to deal with */
                     /* copying data, extending the knot sequence       */
                     /* and a recursive call to this procedure          */
  if ( k < degree ) {
    if ( k < 0 ) {
way_out:
      *fnz = *nnz = 0;
      return;
    }
    auxkn = pkv_GetScratchMemd ( auxsize = 5*(degree+1) );
    auxbfv = &auxkn[2*(degree+1)];
    auxdbfv = &auxbfv[degree+1];
    auxddbfv = &auxdbfv[degree+1];
    l = degree-k;
    memcpy ( &auxkn[l], knots, (2*(degree+1)-l)*sizeof(double) );
    for ( i = 0; i < l; i++ )
      auxkn[i] = knots[0];
    mbs_deBoorBasisDer2d ( degree, 2*degree+1, auxkn, t,
                           fnz, nnz, auxbfv, auxdbfv, auxddbfv );
    if ( *nnz > l ) {
      *nnz -= l;
      memcpy ( bfv, &auxbfv[l], (*nnz)*sizeof(double) );
      memcpy ( dbfv, &auxdbfv[l], (*nnz)*sizeof(double) );
      memcpy ( ddbfv, &auxddbfv[l], (*nnz)*sizeof(double) );
    }
    else
      *nnz = 0;
    *fnz = 0;
    pkv_FreeScratchMemd ( auxsize );
  }
  else if ( k >= lastknot-degree ) {
    if ( r == degree+1 ) {
      *fnz = lastknot-degree-1;  *nnz = 1;
      bfv[0] = 1.0;
      return;
    }
    if ( k >= lastknot ) {
      *fnz = *nnz = 0;
      return;
    }
    auxkn = pkv_GetScratchMemd ( auxsize = 5*(degree+1) );
    auxbfv = &auxkn[2*(degree+1)];
    auxdbfv = &auxbfv[degree+1];
    auxddbfv = &auxdbfv[degree+1];
    l = k+1+degree-lastknot;
    memcpy ( auxkn, &knots[k-degree], (2*(degree+1)-l)*sizeof(double) );
    for ( i = 2*(degree+1)-l; i < 2*(degree+1); i++ )
      auxkn[i] = knots[lastknot];
    mbs_deBoorBasisDer2d ( degree, 2*degree+1, auxkn, t,
                           fnz, nnz, auxbfv, auxdbfv, auxddbfv );
    *fnz = k-degree;
    if ( (*fnz)+(*nnz) > lastknot-degree )
      *nnz = lastknot-degree-(*fnz);
    if ( *nnz > 0 ) {
      memcpy ( bfv, auxbfv, (*nnz)*sizeof(double) );
      memcpy ( dbfv, auxdbfv, (*nnz)*sizeof(double) );
      memcpy ( ddbfv, auxddbfv, (*nnz)*sizeof(double));
    }
    pkv_FreeScratchMemd ( auxsize );
  }
  else {
                     /* now the regular case, */
                     /* with knots[degree] <= t < knots[lastknot-degree] */
    if ( r > degree )
      r = degree;
    *fnz = l = k-degree;
    *nnz = degree+1-r;

    bfv[degree] = 1.0;
    for ( j = 1; j <= degree; j++ ) {
      beta = (knots[k+1]-t)/(knots[k+1]-knots[k-j+1]);
      bfv[k-j-l] = beta*bfv[k-j+1-l];
      for ( i = k-j+1; i < k; i++ ) {
        alpha = 1.0-beta;
        beta = (knots[i+j+1]-t)/(knots[i+j+1]-knots[i+1]);
        bfv[i-l] = alpha*bfv[i-l] + beta*bfv[i+1-l];
      }
      bfv[k-l] = (1.0-beta)*bfv[k-l];
      if ( j == degree-2 )
        memcpy ( ddbfv, bfv, (degree+1)*sizeof(double) );
      if ( j == degree-1 )
        memcpy ( dbfv, bfv, (degree+1)*sizeof(double) );
    }
    alpha = (double)(degree-1)/(knots[k+1]-knots[k-degree+2]);
    beta = alpha*ddbfv[2];
    ddbfv[1] = -beta;
    for ( i = 2; i < *nnz-1; i++ ) {
      alpha = (double)(degree-1)/(knots[k+i]-knots[k-degree+i+1]);
      gamma = alpha*ddbfv[i+1];
      ddbfv[i] = beta-gamma;
      beta = gamma;
    }
    ddbfv[i] = beta;
    alpha = (double)degree/(knots[k+1]-knots[k-degree+1]);
    beta = alpha*dbfv[1];
    dbfv[0] = -beta;
    delta = alpha*ddbfv[1];
    ddbfv[0] = -delta;
    for ( i = 1; i < *nnz-1; i++ ) {
      alpha = (double)degree/(knots[k+i+1]-knots[k-degree+i+1]);
      gamma = alpha*dbfv[i+1];
      dbfv[i] = beta-gamma;
      beta = gamma;
      eps = alpha*ddbfv[i+1];
      ddbfv[i] = delta-eps;
      delta = eps;
    }
    dbfv[i] = beta;
    ddbfv[i] = eps;
  }
} /*mbs_deBoorBasisDer2d*/

