
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"

#undef CONST_
#define CONST_

#include "multibs.h"

/* ///////////////////////////////////////////////////////////////////////// */
static void _mbs_SetupBetaConnectionMatrixf ( int n, CONST_ float *beta,
                                              float *cm )
{
  int i, j, k;

  if ( n < 1 || n > MAXBETADEG )
    return;
        /* compute the first column and diagonal elements */
  cm[0] = 1.0;
  for ( i = 1, j = 0;  i < n;  i++ ) {
    k = j+i+1;
    cm[j+1] = 0.0;
    cm[k] = beta[0]*cm[j];
    j = k;
  }
        /* compute other elements */
  switch ( n ) {
case 5:
    cm[11] = beta[3];
    cm[12] = 3.0*beta[1]*beta[1] + 4.0*beta[0]*beta[2];
    cm[13] = 4.0*beta[0]*beta[0]*beta[1];
case 4:
    cm[7] = beta[2];
    cm[8] = 3.0*beta[0]*beta[1];
case 3:
    cm[4] = beta[1];
case 2:
case 1:
    return;
  }
} /*_mbs_SetupBetaConnectionMatrixf*/

static void _mbs_SetupBetaTransitionMatrixf ( int n, float h_i, float *dm )
{
  int i;

  dm[0] = 1.0;
  for ( i = 1; i < n; i++ )
    dm[i] = dm[i-1]*h_i/(float)i;
} /*_mbs_SetupBetaTransitionMatrixf*/

static void _mbs_MultFf ( int n, float c, float h, float *hatp )
{
  int    k;
  float hk;

  hatp[0] *= c;
  hk = c*h/(float)n;
  for ( k = 1; k <= n; k++ ) {
    hatp[k] *= hk;
    if ( k < n )
      hk *= h/(float)(n-k);
  }
  for ( k = 0; k < n-k; k++ )
    { hk = hatp[k];  hatp[k] = hatp[n-k];  hatp[n-k] = hk; }
} /*_mbs_MultFf*/

boolean mbs_ConstructBetaFuncf ( int n, int lkn, CONST_ float *knots,
                                 CONST_ float *beta,
                                 float *betabez, float *scf )
{
  void   *sp;
  float *cm, *dm, *gi, *g;
  float h[MAXBETADEG], h_i, c_i;
  float ab[MAXBETADEG*MAXBETADEG], x[MAXBETADEG], y[MAXBETADEG+1];
  float e[((MAXBETADEG+1)*(MAXBETADEG+2))/2];
  int    cms;
  int    i, j, k, l, fct;

  sp = pkv_GetScratchMemTop ();
  if ( n < 2 || n > MAXBETADEG || lkn <= 2*n )
    goto failure;
  cms = (n*(n+1))/2;
  cm = pkv_GetScratchMemf ( (lkn-1)*cms );
  dm = pkv_GetScratchMemf ( n*lkn );
  gi = pkv_GetScratchMemf ( (lkn-1)*(n+1)*n );
  if ( !cm || !dm || !gi )
    goto failure;
        /* compute the connection and transition matrices */
  for ( i = 1; i < lkn; i++ ) {
    _mbs_SetupBetaConnectionMatrixf ( n, &beta[(i-1)*(n-1)], &cm[(i-1)*cms] );
    h_i = knots[i+1]-knots[i];
    _mbs_SetupBetaTransitionMatrixf ( n, h_i, &dm[i*n] );
  }
        /* compute the auxiliary functions */
  for ( i = 0; i < lkn; i++ ) {
    g = &gi[i*(n+1)*n];
memset ( g, 0, (n+1)*n*sizeof(float) );
    g[n-1] = h_i = knots[i+1]-knots[i];
    for ( j = n-2; j >= 0; j-- )
      g[j] = g[j+1]*h_i;
    fct = n;
    for ( j = 1, fct = n;  j < n;  j++ ) {
      g[j] *= fct;
      fct *= (n-j);
    }
    for ( j = 1; j <= n; j++ ) {
      if ( i+j == lkn )
        break;
      pkn_LowerTrMatrixMultf ( n, &cm[(i+j-1)*cms], 1, 1, &g[(j-1)*n], 1, h );
      memcpy ( &g[j*n], h, n*sizeof(float) );
      for ( k = 0; k < n-1; k++ )
        g[j*n+k] += pkn_ScalarProductf ( n-1-k, &dm[(i+j)*n+1], &h[k+1] );
    }
  }
        /* construct the beta-spline functions in Bernstein bases */
          /* setup the conversion matrix */
  ab[0] = ab[1] = ab[2] = 1.0;
  for ( i = 2, j = 3; i <= n; i++ ) {
    ab[j] = 1.0;
    for ( k = 1; k < i; k++ )
      ab[j+k] = ab[j+k-i-1]+ab[j+k-i];
    ab[j+i] = 1.0;
    j += i+1;
  }
  for ( i = k = 0;  i <= n;  i++ )
    for ( j = 0;  j <= i;  j++, k++ ) {
      l = pkn_SymMatIndex ( n-j, n-i );
      e[k] = (i+n) % 2 ? -ab[l] : ab[l];
    }
         /* compute fct = n! */
  fct = 2;
  for ( i = 3; i <= n; i++ )
    fct *= i;
          /* now deal with the functions */
  for ( i = 0; i < lkn-n; i++ ) {
          /* assemble the matrix A and solve the system with this matrix */
memset ( ab, 0, n*n*sizeof(float) );
    for ( j = 0; j < n; j++ )
      pkv_Selectf ( n, 1, 1, n, &gi[(i+j+1)*(n+1)*n+n*(n-j-1)], &ab[j] );
    memcpy ( x, &gi[(i+1)*(n+1)*n-n], n*sizeof(float) );
    if ( !pkn_multiGaussSolveLinEqf ( n, ab, 1, 1, x ) )
      goto failure;
/*memset ( x, 0, n*sizeof(float) );*/
          /* assemble the matrix B and solve the system with this matrix */
memset ( ab, 0, n*n*sizeof(float) );
    memcpy ( y, &gi[i*(n+1)*n+n*(n-1)], n*sizeof(float) );
    for ( j = 1; j < n; j++ )
      pkn_AddMatrixMf ( 1, n, 0, y, 0, &gi[(i+j)*(n+1)*n+(n-j-1)*n],
                        -x[j-1], 0, y );
    pkv_Selectf ( n, 1, 1, n, y, ab );
    for ( j = 1; j < n; j++ )
      pkv_Selectf ( n, 1, 1, n, &gi[(i+j)*(n+1)*n+(n-j-1)*n], &ab[j] );
    y[0] = 1.0;
    memset ( &y[1], 0, (n-1)*sizeof(float) );
    if ( !pkn_multiGaussSolveLinEqf ( n, ab, 1, 1, y ) )
      goto failure;
    c_i = y[0];
    if ( scf )
      scf[i] = c_i;
          /* find the representation of polynomials and convert to the */
          /*Bernstein basis */
    if ( i >= n ) {
      memcpy ( ab, &gi[i*(n+1)*n], n*sizeof(float) );
      h_i = knots[i+1]-knots[i];
      ab[n] = fct;
      _mbs_MultFf ( n, c_i, h_i, ab );
      pkn_UpperTrMatrixSolvef ( n+1, e, 1, 1, ab, 1, ab );
      pkv_Selectf ( n+1, 1, 1, n+1, ab,
                    &betabez[(i-n)*(n+1)*(n+1)+n] );
    }
    for ( k = 1; k <= n; k++ ) {
      h_i = knots[i+k+1]-knots[i+k];
      if ( i+k >= n && i+k < lkn-n ) {
        memcpy ( ab, &gi[i*(n+1)*n+k*n], n*sizeof(float) );
        for ( j = 1; j <= k; j++ )
          pkn_AddMatrixMf ( 1, n, 0, ab, 0, &gi[(i+j)*(n+1)*n+(k-j)*n],
                            -x[j-1], 0, ab );
        ab[n] = -x[k-1]*fct;
        _mbs_MultFf ( n, c_i, h_i, ab );
        pkn_UpperTrMatrixSolvef ( n+1, e, 1, 1, ab, 1, ab );
        pkv_Selectf ( n+1, 1, 1, n+1, ab,
                      &betabez[(i+k-n)*(n+1)*(n+1)+n-k] );
      }
    }
  }
        /* compensate for rounding errors */
  cm = betabez;
  for ( i = 0;  i < lkn-2*n;  i++, cm += (n+1)*(n+1) )
    for ( j = 0, k = n;  j < n;  j++, k += n+1 )
      cm[k] = cm[k+1] = 0.0;
  cm = betabez;
  for ( i = 0;  i < lkn-2*n-1;  i++, cm += (n+1)*(n+1) )
    for ( j = 0; j < n; j++ )
      cm[n*(n+1)+1+j] = cm[(n+1)*(n+1)+j] = 0.5*(cm[n*(n+1)+1+j]+cm[(n+1)*(n+1)+j]);

  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_ConstructBetaFuncf*/

/* ///////////////////////////////////////////////////////////////////////// */
boolean mbs_multiGetBetaBezArcf ( int n, int lkn, int ncurves, int spdimen,
                                  CONST_ float *betabez,
                                  int pitch, CONST_ float *cpoints,
                                  int k, float *bezcp )
{
  int i;

  if ( k < 0 || k >= lkn-2*n || ncurves < 1 )
    return false;

  for ( i = 0; i < ncurves; i++ )
    pkn_MultMatrixf ( n+1, n+1, n+1, &betabez[k*(n+1)*(n+1)],
                      spdimen, spdimen, &cpoints[i*pitch+k*spdimen],
                      spdimen, &bezcp[i*(n+1)*spdimen] );
  return true;
} /*mbs_multiGetBetaBezArcf*/

