
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "pklinkage.h"

/* ///////////////////////////////////////////////////////////////////////// */
void pkkl_DefaultTransform ( pkkl_geom_obj *obj, trans3d *tr,
                             int nv, int *vn )
{
  int     i, k;
  double  *vert, *tvert;

  vert = obj->vert;
  tvert = obj->tvert;
  if ( !vert || !tvert )
    return;
  if ( vn ) {
    switch ( obj->nvc ) {
  case 3:
      for ( i = 0; i < nv; i++ ) {
        k = 3*vn[i];
        TransPoint3d ( tr, (point3d*)&vert[k], (point3d*)&tvert[k] );
      }
      break;
  case 4:
      for ( i = 0; i < nv; i++ ) {
        k = 4*vn[i];
        Trans3Point4d ( tr, (point4d*)&vert[k], (point4d*)&tvert[k] );
      }
      break;
  default:
      return;
    }
  }
  else {
    switch ( obj->nvc ) {
  case 3:
      for ( i = k = 0;  i < nv;  i++, k += 3 )
        TransPoint3d ( tr, (point3d*)&vert[k], (point3d*)&tvert[k] );
      break;
  case 4:
      for ( i = k = 0;  i < nv;  i++, k += 4 )
        Trans3Point4d ( tr, (point4d*)&vert[k], (point4d*)&tvert[k] );
      break;
  default:
      return;
    }
  }
} /*pkkl_DefaultTransform*/

void pkkl_obj_redraw_stub ( pkkl_geom_obj *obj, void *usrdata, void *appdata )
{
} /*pkkl_obj_redraw_stub*/

void pkkl_obj_stub ( pkkl_geom_obj *obj )
{
} /*pkkl_obj_stub*/

static char *_pkkl_copyname ( const char *name )
{
  int  lgt;
  char *nc;

  if ( name ) {
    nc = NULL;
    lgt = strlen ( name );
    if ( lgt > 0 ) {
      if ( (nc = malloc ( lgt+1 )) )
        strcpy ( nc, name );
    }
    return nc;
  }
  return NULL;
} /*_pkkl_copyname*/

int pkkl_NewObject ( pkkl_linkage *linkage,
                     int type,
                     const char *name,
                     const char *filename,
                     int nvc, int nvert,
                     const trans3d *etrans,
                     void *data,
                     pkkl_obj_init init,
                     pkkl_obj_transform transform,
                     pkkl_obj_postprocess postprocess,
                     pkkl_obj_redraw redraw,
                     pkkl_obj_destroy destroy )
{
  int           on;
  pkkl_geom_obj *obj;

  if ( linkage->nobj < linkage->maxobj ) {
    on = linkage->nobj;
    obj = &linkage->obj[on];
    memset ( obj, 0, sizeof(pkkl_geom_obj) );
    obj->type = type;
    obj->name = _pkkl_copyname ( name );
    obj->filename = _pkkl_copyname ( filename );
    if ( etrans )
      obj->Etr = *etrans;
    else
      IdentTrans3d ( &obj->Etr );
    obj->transform = transform != NULL ? transform : pkkl_DefaultTransform;
    obj->postprocess = postprocess != NULL ? postprocess : pkkl_obj_stub;
    obj->redraw = redraw != NULL ? redraw : pkkl_obj_redraw_stub;
    obj->destroy = destroy != NULL ? destroy : pkkl_obj_stub;
    obj->nvc = nvc;
    obj->nvert = nvert;
        /* the init procedure must obligatorily initialise the fields */
        /* vert and tvert of the *obj structure */
    if ( init ( obj, data ) ) {
      linkage->nobj ++;
      return on;
    }
    else {
      if ( obj->name )
        free ( obj->name );
      if ( obj->filename )
        free ( obj->filename );
    }
  }
  return -1;
} /*pkkl_NewObject*/

int pkkl_NewLink ( pkkl_linkage *linkage, const char *name )
{
  int       lkn;
  pkkl_link *lk;

  if ( linkage->nlinks < linkage->maxlinks ) {
    lkn = linkage->nlinks ++;
    lk = &linkage->link[lkn];
    memset ( lk, 0, sizeof(pkkl_link) );
    lk->name = _pkkl_copyname ( name );
    lk->fref = lk->fhj = -1;
    IdentTrans3d ( &lk->Mtr );
    lk->tag = 0;
    return lkn;
  }
  return -1;
} /*pkkl_NewLink*/

int pkkl_NewObjRef ( pkkl_linkage *linkage, int lkn,
                     int on, int nv, int *vn )
{
  int          orn;
  pkkl_obj_ref *oref;

  if ( linkage->norefs < linkage->maxorefs ) {
    orn = linkage->norefs ++;
    oref = &linkage->oref[orn];
    memset ( oref, 0, sizeof(pkkl_obj_ref) );
    oref->on = on;
    oref->nv = nv;
    if ( vn ) {
      if ( !(oref->vn = malloc ( nv*sizeof(int) )) )
        return -1;
      memcpy ( oref->vn, vn, nv*sizeof(int) );
    }
    else
      oref->vn = NULL;
    oref->nextr = linkage->link[lkn].fref;
    linkage->link[lkn].fref = orn;
    return orn;
  }
  return -1;
} /*pkkl_NewObjRef*/

int pkkl_NewJoint ( pkkl_linkage *linkage, int l0, int l1, int art, int pnum )
{
  int            jn0, jn1;
  pkkl_halfjoint *hj0, *hj1;

  if ( linkage->nhj < linkage->maxnhj-1 &&
       l0 < linkage->nlinks && l1 < linkage->nlinks ) {
    jn0 = linkage->nhj ++;
    jn1 = linkage->nhj ++;
    hj0 = &linkage->hj[jn0];
    hj1 = &linkage->hj[jn1];
    memset ( hj0, 0, 2*sizeof(pkkl_halfjoint) );
    hj0->otherhalf = jn1;
    hj1->otherhalf = jn0;
    hj0->l0 = hj1->l1 = l0;
    hj0->l1 = hj1->l0 = l1;
    hj0->art = art;
    hj0->pnum = pnum;
    hj1->art = art == PKKL_ART_NONE ? PKKL_ART_NONE : PKKL_ART_OTHERHALF;
    hj1->pnum = -1;
    hj0->nexthj = linkage->link[l0].fhj;
    linkage->link[l0].fhj = jn0;
    hj1->nexthj = linkage->link[l1].fhj;
    linkage->link[l1].fhj = jn1;
    IdentTrans3d ( &hj0->Ftr );
    IdentTrans3d ( &hj0->Atr );
    IdentTrans3d ( &hj0->Btr );
    IdentTrans3d ( &hj1->Ftr );
    IdentTrans3d ( &hj1->Atr );
    IdentTrans3d ( &hj1->Btr );
    return jn0;
  }
  return -1;
} /*pkkl_NewJoint*/

boolean pkkl_obj_init_stub ( pkkl_geom_obj *obj, void *data )
{
  obj->type = PKKL_UNKNOWN;
  obj->vert = obj->tvert = NULL;
  obj->name = obj->filename = NULL;
  IdentTrans3d ( &obj->Etr );
  obj->data0 = obj->data1 = NULL;
  obj->transform = pkkl_DefaultTransform;
  obj->postprocess = pkkl_obj_stub;
  obj->redraw = pkkl_obj_redraw_stub;
  obj->destroy = pkkl_obj_stub;
  return true;
} /*pkkl_obj_init_stub*/

