
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2016, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <sys/times.h>

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

#include "pkbprivate.h"

/* ////////////////////////////////////////////////////////////////////////// */
int _pksbatch_GetNextSymbolFSC ( pks_scanner *sc )
{
  int           ip;
  pks_batchexec *batch;
  char          *code;
  unsigned int  place;
  short int     level;

  batch = sc->libdata;
  batch->fetch_next = false;
  ip = batch->ip;
  code = batch->code;
  place = *(unsigned int*)(&code[ip]);
  ip += sizeof(unsigned int);
  sc->nsline = place / PKS_SCANNER_MAX_LINE_LENGTH;
  sc->nscolumn = place % PKS_SCANNER_MAX_LINE_LENGTH;
  sc->nextsymbol = *(short int*)(&code[ip]);
  ip += sizeof(short int);
  switch ( sc->nextsymbol ) {
case PKS_SYMB_FLOAT:
    sc->nextfloat = *(double*)(&code[ip]);
    ip += sizeof(double);
    break;

case PKS_SYMB_INTEGER:
case PKS_SYMB_CHAR:
case PKS_SYMB_OTHER:
    sc->nextinteger = *(int*)(&code[ip]);
    sc->nextfloat = (double)sc->nextinteger;
    ip += sizeof(int);
    break;

case PKS_SYMB_IDENT:
    sc->id_space = *(short int*)(&code[ip]);
    ip += sizeof(short int);
    sc->nextinteger = *(int*)(&code[ip]);
    ip += sizeof(int);
    strcpy ( sc->nextname, &code[ip] );
    ip += strlen ( sc->nextname )+1;
    break;

case PKS_SYMB_STRING:
    strcpy ( sc->nextname, &code[ip] );
    ip += strlen ( sc->nextname )+1;
    break;

case PKS_SYMB_BATCH_CMD:
    sc->id_space = *(short int*)(&code[ip]);
    ip += sizeof(short int);
    break;

case PKS_SYMB_BATCH_VAR:
    level = *(short int*)(&code[ip]);
    ip += sizeof(short int);
    sc->nextsymbol = PKS_SYMB_INTEGER;
    sc->nextinteger = batch->var[3*level];
    sc->nextfloat = (double)sc->nextinteger;
    break;

default:
    break;
  }
  batch->ip = ip;
  return sc->nextsymbol;
} /*_pksbatch_GetNextSymbolFSC*/

static boolean _pksbatch_IfCMD ( pks_batchexec *batch )
{
  pks_scanner *fsc;
  ifstatement *ifst;
  int         ip;
  pks_calcval x;

  fsc = &batch->fsc;
  ip = batch->ip;
  ifst = (ifstatement*)&batch->code[ip+sizeof(unsigned int)];
        /* evaluate the condition */
  batch->ip += sizeof(ifstatement)+sizeof(unsigned int);
  fsc->GetNextSymbol ( fsc );
  if ( pkscalc_Expression ( fsc, &x ) ) {
    switch ( x.valtype ) {
  case PKS_SYMB_FLOAT:
      x.intval = x.floatval != 0.0;
  case PKS_SYMB_INTEGER:
      batch->ip = x.intval ? ifst->ifptr : ifst->elseptr;
      return true;
  default:
      return false;
    }
  }
  else
    return false;
} /*_pksbatch_IfCMD*/

static boolean _pksbatch_WhileCMD ( pks_batchexec *batch )
{
  pks_scanner *fsc;
  whileloop   *whilest;
  int         ip;
  pks_calcval x;

  fsc = &batch->fsc;
  ip = batch->ip;
  whilest = (whileloop*)&batch->code[ip+sizeof(unsigned int)];
        /* evaluate loop entering condition */
  batch->ip += sizeof(whileloop)+sizeof(unsigned int);
  fsc->GetNextSymbol ( fsc );
  if ( pkscalc_Expression ( fsc, &x ) ) {
    switch ( x.valtype ) {
  case PKS_SYMB_FLOAT:
      x.intval = x.floatval != 0.0;
  case PKS_SYMB_INTEGER:
      batch->ip = x.intval ? whilest->loopptr : whilest->endptr;
      return true;
  default:
      return false;
    }
  }
  else
    return false;
} /*_pksbatch_WhileCMD*/

static boolean _pksbatch_ForCMD ( pks_batchexec *batch, boolean first )
{
  pks_scanner *fsc;
  forloop     *forst;
  int         *var;
  int         ip;
  pks_calcval x;
  boolean     cont;

  fsc = &batch->fsc;
  ip = batch->ip;
  forst = (forloop*)&batch->code[ip+sizeof(unsigned int)];
  var = &batch->var[3*forst->level];
  if ( first ) {  /* first execution - initialise the loop counter */
    batch->ip = forst->initial;
    fsc->GetNextSymbol ( fsc );
    if ( !pkscalc_Expression ( fsc, &x ) )
      goto failure;
    if ( x.valtype != PKS_SYMB_INTEGER )
      goto failure;
    var[0] = x.intval;  /* initial counter value */
    batch->ip = forst->final;
    fsc->GetNextSymbol ( fsc );
    if ( !pkscalc_Expression ( fsc, &x ) )
      goto failure;
    if ( x.valtype != PKS_SYMB_INTEGER )
      goto failure;
    var[1] = x.intval;  /* final counter value */
    batch->ip = forst->incr;
    fsc->GetNextSymbol ( fsc );
    if ( !pkscalc_Expression ( fsc, &x ) )
      goto failure;
    if ( x.valtype != PKS_SYMB_INTEGER )
      goto failure;
    var[2] = x.intval;  /* increment */
    if ( !var[2] )
      var[2] = var[0] < var[1] ? +1 : -1;
  }
  else            /* continuation - add increment to the loop counter */
    var[0] += var[2];
        /* evaluate loop entering condition */
  cont = var[2] > 0 ? var[0] <= var[1] : var[0] >= var[1];
  batch->ip = cont ? forst->loopptr : forst->endptr;
  return true;

failure:
  return false;
} /*_pksbatch_ForCMD*/

static boolean _pksbatch_ContinueCMD ( pks_batchexec *batch )
{
  int           ip, dst;
  keystatement  *keyst;
  jumpstatement *jumpst;

  ip = batch->ip;
  jumpst = (jumpstatement*)&batch->code[ip+sizeof(unsigned int)];
  batch->ip = dst = jumpst->dest;
  keyst = (keystatement*)&batch->code[dst+sizeof(unsigned int)];
  if ( keyst->token != PKS_SYMB_BATCH_CMD )
    return false;
  switch ( keyst->key ) {
case PKS_BATCHCMD_WHILE:
    return _pksbatch_WhileCMD ( batch );
case PKS_BATCHCMD_FOR:
    return _pksbatch_ForCMD ( batch, false );
default:
    return false;
  }
  return true;
} /*_pksbatch_ContinueCMD*/

static boolean _pksbatch_JumpCMD ( pks_batchexec *batch )
{
  int           ip, dest;
  jumpstatement *jumpst;

  ip = batch->ip;
  jumpst = (jumpstatement*)&batch->code[ip+sizeof(unsigned int)];
  dest = jumpst->dest;
  if ( dest >= 0 && dest < batch->codelgt ) {
    batch->ip = dest;
    return true;
  }
  else
    return false;
} /*_pksbatch_JumpCMD*/

static boolean _pksbatch_SwitchCMD ( pks_batchexec *batch )
{
  pks_scanner     *fsc;
  int             ip, swc;
  switchstatement *switchst;
  switchcases     *swcases;
  int             *casetab, ncases, i, j, k;
  pks_calcval     x;

  fsc = &batch->fsc;
  ip = batch->ip;
  switchst = (switchstatement*)&batch->code[ip+sizeof(unsigned int)];
  swc = switchst->cases;
  swcases = (switchcases*)&batch->code[swc+sizeof(unsigned int)];
  ncases = swcases->ncases;
  casetab = (int*)&batch->code[swc+sizeof(unsigned int)+sizeof(switchcases)];
  batch->ip += sizeof(switchstatement) + sizeof(unsigned int);
  fsc->GetNextSymbol ( fsc );
  if ( pkscalc_Expression ( fsc, &x ) ) {
    if ( x.valtype != PKS_SYMB_INTEGER )
      return false;
        /* binary search */
    for ( i = 0, j = ncases; i < j-1; ) {
      k = (i+j)/2;
      if ( x.intval >= casetab[2*k] )
        i = k;
      else
        j = k;
    }
    batch->ip = x.intval == casetab[2*i] ? casetab[2*i+1] : swcases->defaultloc;
    return true;
  }
  else
    return false;
} /*_pksbatch_SwitchCMD*/

static boolean _pksbatch_SkipFilename ( pks_batchexec *batch )
{
  pks_scanner *fsc;
  int         ip;

  fsc = &batch->fsc;
  ip = batch->ip + sizeof(unsigned int) + sizeof(include_filename);
  fsc->filename = &batch->code[ip];
  batch->ip += sizeof(unsigned int) + sizeof(include_filename) +
               strlen( fsc->filename ) + 1;
  return true;
} /*_pksbatch_SkipFilename*/

static boolean _pksbatch_SkipEndInclude ( pks_batchexec *batch )
{
  pks_scanner *fsc;
  include_end *eoi;
  int         ip;

  fsc = &batch->fsc;
  eoi = (include_end*)&batch->code[batch->ip+sizeof(unsigned int)];
  ip = eoi->namepos + sizeof(unsigned int) + sizeof(include_filename);
  fsc->filename = &batch->code[ip];
  batch->ip += sizeof(unsigned int) + sizeof(include_end);
  return true;
} /*_pksbatch_SkipEndInclude*/

int _pksbatch_GetNextSymbolCSC ( pks_scanner *sc )
{
  pks_batchexec *batch;
  pks_scanner   *fsc;

  batch = (pks_batchexec*)sc->libdata;
  fsc = &batch->fsc;
  if ( batch->fetch_next )
    fsc->GetNextSymbol ( fsc );
  for (;;) {
    sc->nsline = fsc->nsline;
    sc->nscolumn = fsc->nscolumn;
    sc->nextsymbol = fsc->nextsymbol;
    switch ( fsc->nextsymbol ) {
  case PKS_SYMB_BATCH_CMD:
      batch->ip -= sizeof(keystatement) + sizeof(unsigned int);
      switch ( fsc->id_space ) {
    case PKS_BATCHCMD_GOTO:
    case PKS_BATCHCMD_JUMP:
        if ( !_pksbatch_JumpCMD ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_CONTINUE:
        if ( !_pksbatch_ContinueCMD ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_IF:
        if ( !_pksbatch_IfCMD ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_WHILE:
        if ( !_pksbatch_WhileCMD ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_FOR:
        if ( !_pksbatch_ForCMD ( batch, true ) )
          goto failure;
        break;
    case PKS_BATCHCMD_SWITCH:
        if ( !_pksbatch_SwitchCMD ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_INCLUDE:
        if ( !_pksbatch_SkipFilename ( batch ) )
          goto failure;
        break;
    case PKS_BATCHCMD_ENDINCL:
        if ( !_pksbatch_SkipEndInclude ( batch ) )
          goto failure;
        break;
    default:
        goto failure;
      }
      break;
  case PKS_SYMB_ERROR:
      goto failure;
  case PKS_SYMB_IDENT:
      strcpy ( sc->nextname, fsc->nextname );
      sc->id_space = fsc->id_space;
      sc->nextinteger = fsc->nextinteger;
      batch->fetch_next = true;
      return sc->nextsymbol;
  case PKS_SYMB_STRING:
      strcpy ( sc->nextname, fsc->nextname );
      batch->fetch_next = true;
      return sc->nextsymbol;
  case PKS_SYMB_INTEGER:
  case PKS_SYMB_CHAR:
      sc->nextinteger = fsc->nextinteger;
      sc->nextfloat = fsc->nextfloat;
      batch->fetch_next = true;
      return sc->nextsymbol;
  case PKS_SYMB_FLOAT:
      sc->nextfloat = fsc->nextfloat;
      batch->fetch_next = true;
      return sc->nextsymbol;
  case PKS_SYMB_OTHER:
      sc->nextinteger = fsc->nextinteger;
      batch->fetch_next = true;
      return sc->nextsymbol;
  default:
      batch->fetch_next = true;
      return sc->nextsymbol;
    }
    fsc->GetNextSymbol ( fsc );
  }

failure:
  return (sc->nextsymbol = PKS_SYMB_ERROR);
} /*_pksbatch_GetNextSymbolCSC*/

/* ////////////////////////////////////////////////////////////////////////// */
void pksbatch_EndExecution ( pks_batchexec *batch )
{
  clock_t ticks;

  ticks = batch->ticks;
  if ( batch->fsc.nextname ) free ( batch->fsc.nextname );
  if ( batch->csc.nextname ) free ( batch->csc.nextname );
  if ( batch->code ) free ( batch->code );
  memset ( batch, 0, sizeof(pks_batchexec) );
  batch->ticks = pkv_Toc ( &ticks );
} /*pksbatch_EndExecution*/

