
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2014, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifndef _SYS_TIMES_H
#include <sys/times.h>
#endif

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

/* ////////////////////////////////////////////////////////////////////////// */
boolean pks_OpenInputFile ( pks_scanner *sc, const char *fn )
{
  pks_file *link;

        /* find out if there is no #include recursion */
  for ( link = sc->infile; link; link = link->inclink )
    if ( !strcmp ( fn, link->filename ) ) {
      sc->nextsymbol = PKS_SYMB_ERROR;
      return false;
    }
        /* open the file */
  PKV_MALLOC ( link, sizeof(pks_file) );
  if ( !link )
    goto failure;
  memset ( link, 0, sizeof(pks_file) );
  PKV_MALLOC ( link->filename, strlen(fn)+1 );
  PKV_MALLOC ( link->inbuffer, sc->inbuflength );
  if ( !link->filename || !link->inbuffer )
    goto failure;
  link->input = fopen ( fn, "r" );
  if ( !link->input )
    goto failure;
  strcpy ( link->filename, fn );
  link->inclink = sc->infile;
  sc->infile = link;
  sc->filename = link->filename;
  if ( sc->IncludeFileHook ) {
    if ( !sc->IncludeFileHook ( sc, true ) )
      goto failure;
  }
  pks_GetNextChar ( sc );
  return true;

failure:
  if ( link ) {
    if ( link->filename ) PKV_FREE ( link->filename );
    if ( link->inbuffer ) PKV_FREE ( link->inbuffer );
    if ( link->input ) fclose ( link->input );
    PKV_FREE ( link );
  }
  return false;
} /*pks_OpenInputFile*/

int pks_ReadInputFile ( pks_scanner *sc )
{
  pks_file *link;

  if ( !(link = sc->infile) )
    goto failure;
  if ( !link->input )
    goto failure;
  link->inbufcount = fread ( link->inbuffer, 1, sc->inbuflength, link->input );
  return link->inbufcount;

failure:
  sc->nextsymbol = PKS_SYMB_ERROR;
  return 0;
} /*pks_ReadInputFile*/

void pks_CloseInputFile ( pks_scanner *sc )
{
  pks_file *link;

  if ( (link = sc->infile) ) {
    if ( sc->IncludeFileHook )
      sc->IncludeFileHook ( sc, false );
    if ( link->input )
      fclose ( link->input );
    if ( link->filename ) PKV_FREE ( link->filename );
    if ( link->inbuffer ) PKV_FREE ( link->inbuffer );
    sc->infile = link->inclink;
    PKV_FREE ( link );
    if ( (link = sc->infile) ) {
      sc->filename = link->filename;
      if ( link->inbufpos >= link->inbufcount ) {
        if ( sc->InputData && link->inbuffer ) {
          sc->InputData ( sc );
          if ( link->inbufcount <= 0 )
            goto end_of_input;
        }
        else {
end_of_input:
          sc->nextchar = link->nextchar = EOF;
          return;
        }
      }
      sc->nextchar = link->nextchar;
    }
    else
      sc->nextchar = EOF;
  }
} /*pks_CloseInputFile*/

int pks_ReadEmptyString ( pks_scanner *sc )
{
  return 0;
} /*pks_ReadEmptyString*/

boolean pks_AppIdentFind ( pks_scanner *sc )
{
  sc->id_space = ID_SPACE_NONE;
  sc->nextinteger = -1;
  return false;
} /*pks_AppIdentFind*/

boolean pks_AppIdentCalc ( pks_scanner *sc, boolean ignore,
                           pks_calcval *value )
{
  return false;
} /*pks_AppIdentCalc*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean pks_InitScanner ( pks_scanner *sc,
                unsigned int options,
                int inbuflength, char *inbuffer, int maxnamelength,
                const char *bcomment, const char *ecomment,
                int (*InputData)(pks_scanner*),
                int (*GetNextSymbol)(pks_scanner*),
                boolean (*AppIdentFind)(pks_scanner*),
                boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                void *userdata )
{
  memset ( sc, 0, sizeof(pks_scanner) );
        /* attach or allocate buffers */
  if ( inbuffer ) {  /* a string will be read in */
    PKV_MALLOC ( sc->infile, sizeof(pks_file) );
    if ( !sc->infile )
      goto failure;
    memset ( sc->infile, 0, sizeof(pks_file) );
    sc->infile->inbuffer = inbuffer;
    sc->infile->inbufcount = strlen(inbuffer);
    sc->alloc_inb = false;
    sc->InputData = pks_ReadEmptyString;
  }
  else {
    if ( inbuflength < 1 )
      inbuflength = PKS_SCANNER_BUF_LENGTH;
    sc->alloc_inb = true;
    sc->InputData = pks_ReadInputFile;
  }
  sc->inbuflength = inbuflength;
  PKV_MALLOC ( sc->nextname, maxnamelength+1 )
  if ( !sc->nextname )
    goto failure;
  sc->namebuflength = maxnamelength+1;
  sc->options = options;
        /* store data */
  if ( options & PKS_OPT_CPP_COMMENTS ) {
    strcpy ( sc->bcomment1, "/*" );
    strcpy ( sc->ecomment1, "*/" );
    strcpy ( sc->bcomment2, "//" );
    strcpy ( sc->ecomment2, "\n" );
  }
  else {
    if ( bcomment ) {
      if ( strlen ( bcomment ) > PKS_MAX_COMMENT_LENGTH )
        goto failure;
      strcpy ( sc->bcomment1, bcomment );
    }
    else sc->bcomment1[0] = 0;
    if ( ecomment ) {
      if ( strlen ( ecomment ) > PKS_MAX_COMMENT_LENGTH )
        goto failure;
      strcpy ( sc->ecomment1, ecomment );
    }
    else sc->ecomment1[0] = 0;
    sc->bcomment2[0] = sc->ecomment2[0] = 0;
  }
        /* default reading procedures may be replaced by some application stuff */
  if ( InputData )
    sc->InputData = InputData;
  if ( GetNextSymbol )
    sc->GetNextSymbol = GetNextSymbol;
  else
    sc->GetNextSymbol = pks_GetNextSymbol;
  if ( AppIdentFind )
    sc->AppIdentFind = AppIdentFind;
  else
    sc->AppIdentFind = pks_AppIdentFind;
  if ( AppIdentCalc )
    sc->AppIdentCalc = AppIdentCalc;
  else
    sc->AppIdentCalc = pks_AppIdentCalc;
  sc->userdata = userdata;
  sc->namebufcount = 0;
  sc->nsline = sc->nscolumn = 0;
  sc->nextsymbol = PKS_SYMB_NONE;
  sc->nrc = 0;
  if ( inbuffer )
    pks_GetNextChar ( sc );
  sc->bkslnl = false;
  return true;

failure:
  pks_ShutDownScanner ( sc );
  return false;
} /*pks_InitScanner*/

void pks_GetNextChar ( pks_scanner *sc )
{
  pks_file *link;

  memmove ( &sc->lc[1], &sc->lc[0], (PKS_MAX_COMMENT_LENGTH-1)*sizeof(short int) );
  sc->lc[0] = sc->nextchar;
  if ( sc->nrc ) {
    sc->nrc --;
    sc->nextchar = sc->rc[sc->nrc];
    if ( (link = sc->infile) )
      link->nextchar = sc->nextchar;
    return;
  }
  if ( !(link = sc->infile) ) {
    sc->nextchar = EOF;
    return;
  }
  if ( link->inbufpos >= link->inbufcount ) {
    if ( !sc->InputData || !link->inbuffer )
      goto end_of_input;
    link->inbufcount = sc->InputData ( sc );
    if ( link->inbufcount <= 0 ) {
end_of_input:
      sc->nextchar = link->nextchar = EOF;
      return;
    }
    link->inbufpos = 0;
  }
  sc->nextchar = link->nextchar = (int)link->inbuffer[link->inbufpos++];
  if ( sc->nextchar == '\n' )
    link->linenum ++, link->colnum = 0;
  else
    link->colnum ++;
} /*pks_GetNextChar*/

static boolean _pks_UngetLastChar ( pks_scanner *sc )
{
  pks_file *link;

  if ( sc->nrc < PKS_MAX_COMMENT_LENGTH ) {
    sc->rc[sc->nrc] = (char)sc->nextchar;
    sc->nextchar = sc->lc[0];
    if ( (link = sc->infile) )
      link->nextchar = sc->nextchar;
    memmove ( &sc->lc[0], &sc->lc[1], (PKS_MAX_COMMENT_LENGTH-1)*sizeof(short int) );
    sc->nrc ++;
    return true;
  }
  else
    return false;
} /*_pks_UngetLastChar*/

boolean _pks_SkipComment ( pks_scanner *sc, const char *bc, const char *ec )
{
  byte rc, lgt;
  char cc[PKS_MAX_COMMENT_LENGTH+1];

  lgt = strlen ( bc );
  if ( lgt < 1 || lgt > PKS_MAX_COMMENT_LENGTH )
    return false;
  for ( cc[0] = (char)sc->nextchar, rc = 1; rc < lgt; rc ++ ) {
    pks_GetNextChar ( sc );
    if ( sc->nextchar == EOF )
      goto unget;
    cc[rc] = (char)sc->nextchar;
  }
  pks_GetNextChar ( sc );
  cc[rc] = 0;
  if ( strcmp ( cc, bc ) )
    goto unget;  /* this is not a comment, unget the characters */
  else {
        /* this is a comment, skip until the comment terminating sequence */
    lgt = strlen ( ec );
    if ( lgt < 1 || lgt > PKS_MAX_COMMENT_LENGTH )
      return false;
    for (;;) {
      while ( sc->nextchar != ec[0] && sc->nextchar != EOF )
        pks_GetNextChar ( sc );
      for ( cc[0] = (char)sc->nextchar, rc = 1; rc < lgt; rc ++ ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == EOF )
          goto unget;
        cc[rc] = (char)sc->nextchar;
      }
      pks_GetNextChar ( sc );
      cc[rc] = 0;
      if ( !strcmp ( cc, ec ) )  /* end of the comment found */
        return true;
      while ( rc > 1 ) {  /* unget all characters but one */
        _pks_UngetLastChar ( sc );
        rc --;
      }
    }
  }

unget:
  while ( rc > 0 ) {
    _pks_UngetLastChar ( sc );
    rc --;
  }
  return false;
} /*_pks_SkipComment*/

int pks_GetNextSymbol ( pks_scanner *sc )
{
  pks_file *link;
#define ADDTONBUF \
{ if ( sc->namebufcount < sc->namebuflength-1 ) \
    sc->nextname[sc->namebufcount++] = (char)sc->nextchar; \
  else \
    return (sc->nextsymbol = PKS_SYMB_ERROR); \
  pks_GetNextChar ( sc ); \
}

  if ( !(link = sc->infile) )
    return (sc->nextsymbol = PKS_SYMB_ERROR);

repeat_scan:
  sc->nsline = link->linenum;
  sc->nscolumn = link->colnum-1;
  sc->nextchar = link->nextchar;
  if ( sc->nextchar == sc->bcomment1[0] ) {
    if ( _pks_SkipComment ( sc, sc->bcomment1, sc->ecomment1 ) )
      goto repeat_scan;
  }
  if ( sc->nextchar == sc->bcomment2[0] ) {
    if ( _pks_SkipComment ( sc, sc->bcomment2, sc->ecomment2 ) )
      goto repeat_scan;
  }
  if ( isdigit ( sc->nextchar ) ) {
        /* integral part */
    sc->nextsymbol = PKS_SYMB_INTEGER;
    sc->namebufcount = 0;
    do {
      ADDTONBUF
    } while ( isdigit ( sc->nextchar ) );
    if ( sc->nextchar == '.' ) {
        /* fractional part */
      ADDTONBUF
      while ( isdigit ( sc->nextchar ) ) {
        ADDTONBUF
      }
      sc->nextsymbol = PKS_SYMB_FLOAT;
    }
    if ( sc->nextchar == 'e' || sc->nextchar == 'E' ) {
        /* exponent */
      ADDTONBUF
      if ( sc->nextchar == '-' || sc->nextchar == '+' ) {
        ADDTONBUF
      }
      if ( !isdigit ( sc->nextchar ) )
        return (sc->nextsymbol = PKS_SYMB_ERROR);
      do {
        ADDTONBUF
      } while ( isdigit ( sc->nextchar ) );
      sc->nextsymbol = PKS_SYMB_FLOAT;
    }
        /* convert the character string to a number */
    sc->nextname[sc->namebufcount] = 0;
    if ( sc->nextsymbol == PKS_SYMB_INTEGER ) {
      sscanf ( sc->nextname, "%d", &sc->nextinteger );
      sc->nextfloat = (double)sc->nextinteger;
    }
    else
      sscanf ( sc->nextname, "%lf", &sc->nextfloat );
  }
  else if ( isalpha ( sc->nextchar ) || sc->nextchar == '_' ) {
        /* an identifier or keyword */
    sc->namebufcount = 0;
    do {
      ADDTONBUF
    } while ( isalpha ( sc->nextchar ) || isdigit (sc->nextchar) ||
              sc->nextchar == '_' );
    sc->nextname[sc->namebufcount] = 0;
    sc->nextsymbol = PKS_SYMB_IDENT;
    sc->AppIdentFind ( sc );
  }
  else {
    switch ( sc->nextchar ) {
  case '\n':
      pks_GetNextChar ( sc );
      if ( sc->bkslnl ) {
        sc->nextsymbol = PKS_SYMB_EOLN;
        goto get_next_char;
      }
      else
        goto repeat_scan;
  case ' ':
  case '\r':
  case '\t':
      pks_GetNextChar ( sc );
      goto repeat_scan;
  case '"':
      pks_GetNextChar ( sc );
      sc->namebufcount = 0;
      do {
        ADDTONBUF
      } while ( sc->nextchar != '"' );
      sc->nextname[sc->namebufcount] = 0;
      pks_GetNextChar ( sc );
      sc->nextsymbol = PKS_SYMB_STRING;
      break;
  case EOF:
      sc->nextsymbol = PKS_SYMB_EOF;
      break;
  case '+':
      sc->nextsymbol = PKS_SYMB_PLUS;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
    case '+': sc->nextsymbol = PKS_SYMB_PLUSPLUS;  break;
    case '=': sc->nextsymbol = PKS_SYMB_PLUSEQ;    break;
    default:  goto next_char_ready;
        }
      }
      goto get_next_char;
  case '-':
      sc->nextsymbol = PKS_SYMB_MINUS;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
    case '-': sc->nextsymbol = PKS_SYMB_MINUSMINUS;  break;
    case '>': sc->nextsymbol = PKS_SYMB_RIGHTARROW;  break;
    case '=': sc->nextsymbol = PKS_SYMB_MINUSEQ;     break;
    default:  goto next_char_ready;
        }
      }
      goto get_next_char;
  case '*':
      sc->nextsymbol = PKS_SYMB_STAR;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_STAREQ;
                              else break;
      }
      goto get_next_char;
  case '/':
      sc->nextsymbol = PKS_SYMB_SLASH;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_SLASHEQ;
                              else break;
      }
      goto get_next_char;
  case '(':
      sc->nextsymbol = PKS_SYMB_LPAREN;    goto get_next_char;
  case ')':
      sc->nextsymbol = PKS_SYMB_RPAREN;    goto get_next_char;
  case '[':
      sc->nextsymbol = PKS_SYMB_LBRACKET;  goto get_next_char;
  case ']':
      sc->nextsymbol = PKS_SYMB_RBRACKET;  goto get_next_char;
  case '{':
      sc->nextsymbol = PKS_SYMB_LBRACE;    goto get_next_char;
  case '}':
      sc->nextsymbol = PKS_SYMB_RBRACE;    goto get_next_char;
  case '<':
      sc->nextsymbol = PKS_SYMB_LANGLE;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
      case '=': sc->nextsymbol = PKS_SYMB_LEQ;  goto get_next_char;
      case '<':
          sc->nextsymbol = PKS_SYMB_LLANGLE;
          pks_GetNextChar ( sc );
          if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_LLANGLEEQ;
                                else goto next_char_ready;
          break;
      default: goto next_char_ready;
        }
      }
      goto get_next_char;
  case '>':
      sc->nextsymbol = PKS_SYMB_RANGLE;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
      case '=': sc->nextsymbol = PKS_SYMB_GEQ;  goto get_next_char;
      case '>':
          sc->nextsymbol = PKS_SYMB_RRANGLE;
          pks_GetNextChar ( sc );
          if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_RRANGLEEQ;
                                else goto next_char_ready;
          break;
      default: goto next_char_ready;
        }
      }
      goto get_next_char;
  case '=':
      sc->nextsymbol = PKS_SYMB_EQ;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_EQEQ;
                              else break;
      }
      goto get_next_char;
  case '\'':
      sc->nextsymbol = PKS_SYMB_CHAR;
      pks_GetNextChar ( sc );
      sc->nextinteger = sc->nextchar;
      pks_GetNextChar ( sc );
      if ( sc->nextchar != '\'' )
        { sc->nextsymbol = PKS_SYMB_ERROR;  break; }
      else
        goto get_next_char;
  case ',':
      sc->nextsymbol = PKS_SYMB_COMMA;     goto get_next_char;
  case '.':
      sc->nextsymbol = PKS_SYMB_DOT;       goto get_next_char;
  case ';':
      sc->nextsymbol = PKS_SYMB_SEMICOLON; goto get_next_char;
  case ':':
      sc->nextsymbol = PKS_SYMB_COLON;     goto get_next_char;
  case '%':
      sc->nextsymbol = PKS_SYMB_PERCENT;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_PERCENTEQ;
                              else break;
      }
      goto get_next_char;
  case '#':
      sc->nextsymbol = PKS_SYMB_HASH;      goto get_next_char;
  case '!':
      sc->nextsymbol = PKS_SYMB_EXCLAM;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_NEQ;
                              else break;
      }
      goto get_next_char;
  case '~':
      sc->nextsymbol = PKS_SYMB_TILDE;     goto get_next_char;
  case '@':
      sc->nextsymbol = PKS_SYMB_AT;        goto get_next_char;
  case '$':
      sc->nextsymbol = PKS_SYMB_DOLLAR;    goto get_next_char;
  case '&':
      sc->nextsymbol = PKS_SYMB_ET;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
      case '&': sc->nextsymbol = PKS_SYMB_ETET;  break;
      case '=': sc->nextsymbol = PKS_SYMB_ETEQ;  break;
      default:  goto next_char_ready;
        }
      }
      goto get_next_char;
  case '\\':
      if ( sc->bkslnl ) {
        do {
          pks_GetNextChar ( sc );
        } while ( sc->nextchar != '\n' && sc->nextchar != EOF );
        pks_GetNextChar ( sc );
        goto repeat_scan;
      }
      else {
        sc->nextsymbol = PKS_SYMB_BACKSLASH;
        goto get_next_char;
      }
  case '?':
      sc->nextsymbol = PKS_SYMB_QUESTION;  goto get_next_char;
  case '|':
      sc->nextsymbol = PKS_SYMB_VERTBAR;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        switch ( sc->nextchar ) {
      case '|': sc->nextsymbol = PKS_SYMB_VERTBARBAR;  break;
      case '=': sc->nextsymbol = PKS_SYMB_VERTBAREQ;   break;
      default:  goto next_char_ready;
        }
      }
      goto get_next_char;
  case '^':
      sc->nextsymbol = PKS_SYMB_WEDGE;
      if ( sc->options & PKS_OPT_C_OPERATORS ) {
        pks_GetNextChar ( sc );
        if ( sc->nextchar == '=' ) sc->nextsymbol = PKS_SYMB_WEDGEEQ;
                              else break;
      }
      goto get_next_char;
  default:
      sc->nextsymbol = PKS_SYMB_OTHER;
      sc->nextinteger = sc->nextchar;
get_next_char:
      pks_GetNextChar ( sc );
      break;
    }
  }
next_char_ready:
  return sc->nextsymbol;
#undef ADDTONBUF
} /*pks_GetNextSymbol*/

void pks_ShutDownScanner ( pks_scanner *sc )
{
  pks_file *link;

  while ( sc->infile ) {
    link = sc->infile;
    sc->infile = link->inclink;
    if ( sc->alloc_inb && link->inbuffer )
      PKV_FREE ( link->inbuffer );
    if ( link->filename )
      PKV_FREE ( link->filename );
    if ( link->input )
      fclose ( link->input );
    PKV_FREE ( link );
  }
  if ( sc->nextname ) PKV_FREE ( sc->nextname );
  memset ( sc, 0, sizeof(pks_scanner) );
  sc->nextsymbol = PKS_SYMB_ERROR;
} /*pks_ShutDownScanner*/

int pks_BinSearchNameTable ( int nnames, int firstname, const char **names,
                             const char *name )
{
  int i, j, k, l;

  i = k = 0;  j = nnames;
  do {
    k = (i+j)/2;
    l = strcmp ( name, names[k] );
    if ( l == 0 )
      return firstname + k;
    else if ( l < 0 )
      j = k;
    else
      i = k+1;
  } while ( j-i > 0 );
  return PKS_SYMB_ERROR;
} /*pks_BinSearchNameTable*/

