
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2008, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

#include "pkvaria.h"
#include "xgledit.h"
#include "xgeprivate.h"
#include "xgleprivate.h"

#define MAX_GLX_FB_ATTRIB 20
#define MAX_GL_CTX_ATTRIB  5


/* ////////////////////////////////////////////////////////////////////////// */
/* The method of creating an OpenGL context for versions 3.0 or above, */
/* implemented below, is based on the sample code from the tutorial at */
/* https://www.khronos.org/opengl/wiki/Tutorial:_OpenGL_3.0_Context_Creation_(GLX)*/

typedef GLXContext (*glXCreateContextAttribsARBProc)
                       (Display*, GLXFBConfig, GLXContext, boolean, const int*);

static boolean ctxErrorOccurred = false;
int (*oldHandler)(Display*, XErrorEvent*);

/* Helper to check for extension string presence. Adapted from: */
/*   http://www.opengl.org/resources/features/OGLextensions/ */
boolean xgle_IsExtensionSupported ( const char *extList, const char *extension )
{
  const char *start;
  const char *where, *terminator;
  
  /* Extension names should not have spaces. */
  where = strchr ( extension, ' ' );
  if ( where || *extension == '\0' )
    return false;

  /* It takes a bit of care to be fool-proof about parsing the
     OpenGL extensions string. Don't be fooled by sub-strings, etc. */
  for ( start = extList; ; ) {
    where = strstr ( start, extension );
    if ( !where )
      break;
    terminator = where + strlen ( extension );
    if ( where == start || *(where - 1) == ' ' )
      if ( *terminator == ' ' || *terminator == '\0' )
        return true;
    start = terminator;
  }
  return false;
} /*xgle_IsExtensionSupported*/

static int ctxErrorHandler ( Display *dpy, XErrorEvent *ev )
{
  ctxErrorOccurred = true;
  return 0;
} /*ctxErrorHandler*/

static void _xgle_InitFBAttr ( unsigned int mask, int bits, int *attribs )
{
  int n;

  n = 0;
  attribs[n++] = GLX_X_RENDERABLE,     attribs[n++] = true;
  attribs[n++] = GLX_DRAWABLE_TYPE,    attribs[n++] = GLX_WINDOW_BIT;
  attribs[n++] = GLX_RENDER_TYPE,      attribs[n++] = GLX_RGBA_BIT;
  attribs[n++] = GLX_X_VISUAL_TYPE,    attribs[n++] = GLX_TRUE_COLOR;
  attribs[n++] = GLX_RED_SIZE,         attribs[n++] = bits;
  attribs[n++] = GLX_RED_SIZE,         attribs[n++] = bits;
  attribs[n++] = GLX_RED_SIZE,         attribs[n++] = bits;
  attribs[n++] = GLX_ALPHA_SIZE,       attribs[n++] = bits;
  attribs[n++] = GLX_DOUBLEBUFFER,     attribs[n++] = false;
  attribs[n++] = GLX_DRAWABLE_TYPE,    attribs[n++] = GLX_WINDOW_BIT | GLX_PIXMAP_BIT;
  if ( mask & XGLE_DEPTH_BUFFER )
    attribs[n++] = GLX_DEPTH_SIZE,       attribs[n++] = 24;
  if ( mask & XGLE_ACCUM_BUFFER ) {
    attribs[n++] = GLX_ACCUM_RED_SIZE,   attribs[n++] = 8;
    attribs[n++] = GLX_ACCUM_GREEN_SIZE, attribs[n++] = 8;
    attribs[n++] = GLX_ACCUM_BLUE_SIZE,  attribs[n++] = 8;
    attribs[n++] = GLX_ACCUM_ALPHA_SIZE, attribs[n++] = 8;
  }
  if ( mask & XGLE_STENCIL_BUFFER )
    attribs[n++] = GLX_STENCIL_SIZE,     attribs[n++] = 8;
  attribs[n] = None;
} /*_xgle_InitFBAttr*/

static void _xgle_InitCTXAttr ( int majorgl, int minorgl,
                                int *attribs )
{
  int n;

  n = 0;
  attribs[n++] = GLX_CONTEXT_MAJOR_VERSION_ARB, attribs[n++] = majorgl;
  attribs[n++] = GLX_CONTEXT_MINOR_VERSION_ARB, attribs[n++] = minorgl;
  if ( majorgl >= 3 ) {
    attribs[n++] = GLX_CONTEXT_FLAGS_ARB,
    attribs[n++] = GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB;
  }
  attribs[n++] = GLX_CONTEXT_PROFILE_MASK_ARB,
  attribs[n++] = majorgl >= 3 ? GLX_CONTEXT_CORE_PROFILE_BIT_ARB
                              : GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB;
  attribs[n] = None;
} /*_xgle_InitCTXAttr*/

static boolean _xge_ChooseFBConfig ( Display *dpy,
                                     int majorgl, int minorgl,
                                     unsigned int required, unsigned int nice,
                                     GLXFBConfig *bestfbc, XVisualInfo **vi )
{
  int        attribs[2*MAX_GLX_FB_ATTRIB+1];
  GLXFBConfig *fbc;
  XVisualInfo *vii;
  int         glx_major, glx_minor;
  int         best_fbc = -1, worst_fbc = -1,
              best_num_samp = -1, worst_num_samp = 999;
  int         samp_buf, samples;
  int         i, fbcount;

  /* FBConfigs were added in GLX version 1.3.*/
  if ( !glXQueryVersion ( dpy, &glx_major, &glx_minor ) || 
       ( ( glx_major == 1 ) && ( glx_minor < 3 ) ) || ( glx_major < 1 ) ) {
    printf ( "Invalid GLX version\n" );
    return false;
  }
/*  printf ( "Getting matching framebuffer configs\n" );*/
  nice |= required;
        /* negotiate OpenGL resources */
  _xgle_InitFBAttr ( nice, 8, attribs );
  fbc = glXChooseFBConfig ( dpy, DefaultScreen(dpy), attribs, &fbcount );
  if ( fbc ) {
    _xgle_depth   = (nice & XGLE_DEPTH_BUFFER) != 0;
    _xgle_accum   = (nice & XGLE_ACCUM_BUFFER) != 0;
    _xgle_stencil = (nice & XGLE_STENCIL_BUFFER) != 0;
  }
  else {
    _xgle_InitFBAttr ( required, 5, attribs );
    fbc = glXChooseFBConfig ( dpy, DefaultScreen(dpy), attribs, &fbcount );
    if ( fbc ) {
      _xgle_depth   = (required & XGLE_DEPTH_BUFFER) != 0;
      _xgle_accum   = (required & XGLE_ACCUM_BUFFER) != 0;
      _xgle_stencil = (required & XGLE_STENCIL_BUFFER) != 0;
    }
    else {
      printf( "Failed to retrieve a framebuffer config\n" );
      return false;
    }
  }
  /* Pick the FB config/visual with the most samples per pixel*/
  for ( i = 0; i < fbcount; ++i ) {
    vii = glXGetVisualFromFBConfig ( dpy, fbc[i] );
    if ( vii ) {
      glXGetFBConfigAttrib ( dpy, fbc[i], GLX_SAMPLE_BUFFERS, &samp_buf );
      glXGetFBConfigAttrib ( dpy, fbc[i], GLX_SAMPLES       , &samples  );
/*
      printf( "  Matching fbconfig %d, visual ID 0x%2x: SAMPLE_BUFFERS = %d,"
              " SAMPLES = %d\n", 
              i, vii -> visualid, samp_buf, samples );
*/
      if ( best_fbc < 0 || (samp_buf && samples > best_num_samp) )
        best_fbc = i, best_num_samp = samples;
      if ( worst_fbc < 0 || !samp_buf || samples < worst_num_samp )
        worst_fbc = i, worst_num_samp = samples;
    }
    XFree( vii );
  }
  *bestfbc = fbc[best_fbc];
  /* Be sure to free the FBConfig list allocated by glXChooseFBConfig()*/
  XFree ( fbc );
  /* Get a visual*/
  *vi = glXGetVisualFromFBConfig ( dpy, *bestfbc );
/*  printf( "Chosen visual ID = 0x%x\n", (unsigned int)(*vi)->visualid );*/
  return true;
} /*_xge_ChooseFBConfig*/

void xgle_ExitOnGLError ( const char *msg )
{
  GLenum err;

  if ( (err = glGetError ()) != GL_NO_ERROR ) {
    fprintf ( stderr, "Error %d: %s, %s\n", err, gluErrorString ( err ), msg );
    exit ( 1 );
  }
} /*xgle_ExitOnGLError*/

void xgle_IgnoreGLErrors ( boolean output )
{
  GLenum err;

  while ( (err = glGetError ()) != GL_NO_ERROR ) {
    if ( output )
      fprintf ( stderr, "Error %d: %s\n", err, gluErrorString ( err ) );
  }
} /*xgle_IgnoreGLErrors*/

boolean _xgle_InitGLXContext ( int majorgl, int minorgl,
                               unsigned int required, unsigned int nice )
{
  glXCreateContextAttribsARBProc glXCreateContextAttribsARB;
  const char                     *glxExts;
  GLXFBConfig                    bestfbc;
  int                            attribs[2*MAX_GL_CTX_ATTRIB+1];

  if ( !glXQueryExtension ( xgedisplay, NULL, NULL ) ) {
    printf ( "Error: GL X extension not supported\n" );
    return false;
  }
  if ( !_xge_ChooseFBConfig ( xgedisplay, majorgl, minorgl, required, nice,
                              &bestfbc, &xgevisualinfo ) )
    return false;
  xgevisual = xgevisualinfo->visual;
  if ( xgevisual->class != TrueColor ) {
    printf ( "Error: True colour visual is needed." );
    return false;
  }
  xge_nplanes = XDisplayPlanes ( xgedisplay, xgescreen );
  xgecolormap = XCreateColormap( xgedisplay,
                      DefaultRootWindow (xgedisplay), 
                      xgevisual, AllocNone );
  if ( !xgecolormap ) {
    printf ( "Error: Cannot create colormap\n" );
    return false;
  }
  /* Get the default screen's GLX extension list*/
  glxExts = glXQueryExtensionsString ( xgedisplay, DefaultScreen( xgedisplay ) );
  /* NOTE: It is not necessary to create or make current to a context before */
  /* calling glXGetProcAddressARB*/
  glXCreateContextAttribsARB = (glXCreateContextAttribsARBProc)
           glXGetProcAddress ( (const GLubyte*)"glXCreateContextAttribsARB" );

  /* Install an X error handler so the application won't exit if GL >= 3.0 */
  /* context allocation fails.*/
  ctxErrorOccurred = false;
  oldHandler = XSetErrorHandler ( &ctxErrorHandler );

  /* Check for the GLX_ARB_create_context extension string and the function.*/
  /* If either is not present, use GLX 1.3 context creation method.*/
  if ( majorgl < 3 )
    xglecontext = glXCreateNewContext ( xgedisplay, bestfbc, GLX_RGBA_TYPE, 0, true );
  else if ( !xgle_IsExtensionSupported ( glxExts, "GLX_ARB_create_context" ) ||
            !glXCreateContextAttribsARB )
    return false;
  else {  /* try to get a GL 3.0 or higher context */

{int err;
if ( (err = glGetError ()) != GL_NO_ERROR )
  fprintf ( stderr, "Error %d: %s, qq 0\n", err, gluErrorString ( err ) );
}
    _xgle_InitCTXAttr ( majorgl, minorgl, attribs );
    xglecontext = glXCreateContextAttribsARB ( xgedisplay, bestfbc, 0,
                                               true, attribs );
{int err;
if ( (err = glGetError ()) != GL_NO_ERROR )
  fprintf ( stderr, "Error %d: %s, qq 1\n", err, gluErrorString ( err ) );
}
    /* Sync to ensure any errors generated are processed.*/
    XSync ( xgedisplay, false );
    if ( ctxErrorOccurred ) {
      printf( "Error: Failed to create GL %d.%d context\n", majorgl, minorgl );
      return false;
    }
  }
  /* Sync to ensure any errors generated are processed.*/
  XSync ( xgedisplay, false );
  /* Restore the original error handler*/
  XSetErrorHandler ( oldHandler );

  if ( ctxErrorOccurred || !xglecontext ) {
    printf ( "Failed to create an OpenGL context\n" );
    return false;
  }

  glGetIntegerv ( GL_MAJOR_VERSION, &majorgl );
  glGetIntegerv ( GL_MINOR_VERSION, &minorgl );
  _xgle_majorgl = (byte)majorgl;  _xgle_minorgl = (byte)minorgl;
  _xgle_ctx_direct = glXIsDirect ( xgedisplay, xglecontext );
{int err;
if ( (err = glGetError ()) != GL_NO_ERROR )
  fprintf ( stderr, "Error %d: %s, qq 2\n", err, gluErrorString ( err ) );
}
  return true;
} /*_xgle_InitGLXContext*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean xgle_Init ( int argc, char *argv[],
                    int (*callback)(xge_widget*,int,int,short,short),
                    char *title,
                    int majorgl, int minorgl,
                    unsigned int required, unsigned int nice )
{
  if ( !_xge_InitInit ( argc, argv ) )
    return false;
  xge_callback = callback;

  if ( !_xgle_InitGLXContext ( majorgl, minorgl, required, nice ) ) {
    printf ( "Error: Could not create OpenGL context\n" );
    return false;
  }

  _xge_InitWinHints ();
  if ( !title )
    title = xge_p_name;
  xge_NewWindow ( title );
  xge_SetWindow ( 0 );
  xglepixmap = XCreatePixmap ( xgedisplay, xgewindow,
                               xge_MAX_WIDTH, xge_MAX_HEIGHT, xge_nplanes );
  if ( !xglepixmap ) {
    printf ( "Error: Cannot create X pixmap\n" );
    return false;
  }
  _xglepixmap = glXCreateGLXPixmap ( xgedisplay, xgevisualinfo, xglepixmap );
  if ( !_xglepixmap ) {
    printf ( "Error: Cannot create GLX pixmap\n" );
    return false;
  }
  if ( !glXMakeCurrent ( xgedisplay, _xglepixmap, xglecontext ) ) {
    printf ( "Error: glXMakeCurrent failed\n" );
    return false;
  }
  if ( majorgl >= 3 )
    _xgle_GetGLProcedures ();

  _xge_InitGCandCursors ();
  xge_GetWindowSize ();
  _xge_FindAspect ();
  _xgle_MakePalette ();

  _xge_FinishInit ();           
  return true;
} /*xgle_Init*/

/* ////////////////////////////////////////////////////////////////////////// */
void xgle_Cleanup ( void )
{
  int i;

#ifdef USE_XEXT_SHAPE
  _xge_DestroySpecialWin ();
#endif
  XFreeGC ( xgedisplay, xgegc );
  for ( i = xge_winnum-1; i >= 0; i-- ) {
    xge_SetWindow ( i );
    XFreePixmap ( xgedisplay, xgepixmap );
    XDestroyWindow ( xgedisplay, xgewindow );
  }
  glXDestroyContext ( xgedisplay, xglecontext );
  XCloseDisplay ( xgedisplay );
} /*xgle_Cleanup*/

